/*
 *  $Id$
 */

/**
 *  @file bt.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "util.h"
#include "bt_discover.h"
#include "node_protocol.h"
#include "client_protocol.h"
#include "discover_protocol.h"
#include "event.h"
#include "bt_net.h"
#include "sdp.h"

/**
 *  @brief accept incoming bluetooth connections
 *
 *  Accept incoming bluetooth connections, setup/init the
 *  connection. Send notifications to other modules.
 *
 *  @returns -1 on error, 1 on success and 2 with penning or generated events
 */
int bt_server(struct btchatd_t *globals, struct pollfd *fd)
{
	struct sockaddr_rc remote_sa = {0};
	int remote_len = 0;
	int remote_socket = -1;
	int ret_val;
		

	#ifdef DEBUG_BT
	assert(globals != NULL);
	#endif

	
	/* init btserver */
	if (globals->btserver.state == 0) {
		if ((globals->btserver.server = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
			#ifdef DEBUG_BT
			perror("bt_server(): socket");
			#endif
			return(-1);
		}
				
		globals->btserver.server_sa.rc_family = AF_BLUETOOTH;
		globals->btserver.server_sa.rc_channel = globals->self.btchat_channel;
		globals->btserver.server_sa.rc_bdaddr = globals->self.addr;
		
		if (bind(globals->btserver.server, (struct sockaddr*) &globals->btserver.server_sa, sizeof(globals->btserver.server_sa)) < 0) {
			#ifdef DEBUG_BT
			perror("bt_server(): bind");
			#endif
			return(-1);
		}
		
		if (listen(globals->btserver.server, 1) < 0) {
			#ifdef DEBUG_BT
			perror("bt_server(): listen");
			#endif
			return(-1);
		}
	
		/* dynamic channel allocation - get actual channel */
		if (globals->self.btchat_channel == 0) {
			remote_len = sizeof(struct sockaddr_rc);
			memset(&remote_sa, 0, sizeof(struct sockaddr_rc));
			if (getsockname(globals->btserver.server, (struct sockaddr*) &remote_sa, &remote_len) < 0) {
				#ifdef DEBUG_BT
				perror("bt_server(): getsockname()");
				#endif
				return(-1);
			}
			
			/* broken kernel! */
			if (remote_sa.rc_channel == 0) {
				globals->self.btchat_channel = 23;
				globals->btserver.state = 0;
				close(globals->btserver.server);
				
				#ifdef DEBUG_BT_INFO
				printf("bt_server(): fallback to channel %d\n", globals->self.btchat_channel);
				#endif
				
				/* retry bind/listen */
				return(bt_server(globals, NULL));
			}
			/* working kernel */
			else {
				globals->self.btchat_channel = remote_sa.rc_channel;
			
				#ifdef DEBUG_BT_INFO
				printf("bt_server(): listening on channel %d\n", remote_sa.rc_channel);
				#endif
			}
		}		
		
		/* register btChat with just allocated channel with the SDP server */
		bt_sdp_register(globals);
				
		globals->btserver.state = 1;
		return(1);
	} /* state == 0 */
	/* ready to accept new connections */
	else if (globals->btserver.state == 1) {
	
		/* check polled events */
		if ((fd->revents & POLLIN) != POLLIN && (fd->revents & POLLPRI) != POLLPRI) return(1);
	
		if (fd->fd == globals->btserver.server) {
			remote_len = sizeof(struct sockaddr_rc);
			memset(&remote_sa, 0, sizeof(struct sockaddr_rc));
			remote_socket = accept(globals->btserver.server, (struct sockaddr*) &remote_sa, &remote_len);
			
			if (remote_socket != -1) {
				
				/* only allow connection if we are below the limit */
				if (globals->btserver.clients_max <= globals->btserver.clients_num) {
					#ifdef DEBUG_BT_INFO
					printf("bt_server(): connections - current = %d max = %d, too many!\n", globals->btserver.clients_num, globals->btserver.clients_max);
					#endif
					
					close(remote_socket);
					return(1);
				}
				
				/* check remote device and setup connection */
				ret_val = bt_client_direct_incoming(globals, remote_socket, &remote_sa, remote_len);
				return(ret_val);
			}
			else {
				#ifdef DEBUG_BT
				perror("bt_server(): accept\n");
				#endif
				return(-1);
			}
		}
		else {
			#ifdef DEBUG_BT
			printf("bt_server(): fd.fd has strange value!\n");
			#endif
			
			/* set state to self init */
			globals->btserver.state = 0;
			bt_sdp_unregister(globals);
			/* indicate penning events */
			return(2);
		}
	} /* state == 1 */
	/* stop accepting new connections */
	else if (globals->btserver.state == 2) {
		if (globals->btserver.server != -1) {
			close(globals->btserver.server);
			globals->btserver.server = -1;
		}
	}
	
	return(1);
}

/**
 *  @brief handle client connection
 *
 *  @param globals	pointer to globals
 *  @param fd	pointer to poll event
 *  @param index	index of client connection in btserver.clients[X] array
 *
 *  @returns -1 on error 1 on success and 2 on pending or generated events
 */
int bt_client(struct btchatd_t *globals, struct pollfd *fd, int index)
{
	int ret_val = 0;
	int read_bytes = 0;
	int write_bytes = 0;
	struct btchat_event_t *event = NULL;
	struct btcnode_protocol_header_t *he;
	struct btcnode_protocol_route_t *route;
	struct btcnode_protocol_route_full_t *routefull;
	struct btcnode_protocol_msg_t *msg;
	struct btcnode_protocol_trust_t *trust;
	#ifdef DEBUG_BT_INFO
	char bdstr[20];
	#endif
	

	#ifdef DEBUG_BT
	assert(globals != NULL);
	#endif
	

	if (globals->btserver.clients[index].state == 0) {
		#ifdef DEBUG_BT
		printf("bt_client[%d](): state = 0, should not happen!\n", index);
		#endif
	}
	else if (globals->btserver.clients[index].state == 1) {
		if (fd->fd == globals->btserver.clients[index].s) {

			if ((fd->revents & POLLERR) == POLLERR) {
				#ifdef DEBUG_BT_INFO
				printf("bt_client[%d](): poll() indicated ERR, disconnect!\n", index);
				#endif
				
				goto disconnect;
			}
			else if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {				
				read_bytes = cread(globals->btserver.clients[index].s, globals->btserver.clients[index].buffer + globals->btserver.clients[index].buffer_pos, globals->btserver.clients[index].buffer_size - globals->btserver.clients[index].buffer_pos);

				if (read_bytes == 0) {
					#ifdef DEBUG_BT_INFO
					printf("bt_client[%d](): other side disconnected\n", index);
					#endif
					
					goto disconnect;
				}
				/* advance buffer position */
				globals->btserver.clients[index].buffer_pos += read_bytes;

				/* increment received statistic counter */
				globals->btserver.clients[index].bytes_recv += read_bytes;

			check_buffer_again:
				if (globals->btserver.clients[index].buffer_pos >= sizeof(struct btcnode_protocol_header_t)) {
					he = (struct btcnode_protocol_header_t*) globals->btserver.clients[index].buffer;
				
					/* change byte order network->host */
					he->time = ntohl(he->time);
					he->length = ntohs(he->length);
				
					#ifdef DEBUG_BT_INFO
					printf("bt_client[%d](): he->length = %d, buffer_pos = %d\n", index, he->length, globals->btserver.clients[index].buffer_pos);
					#endif
					
					/* broken packet - reset buffer */
					if (he->length == 0) globals->btserver.clients[index].buffer_pos = 0;
				
					if (globals->btserver.clients[index].buffer_pos >= he->length) {
			
						/* trust level notification */
						if (he->pkg_type == btcnode_protocol_trust) {
							#ifdef DEBUG_BT_INFO
							printf("bt_client[%d](): trust packet\n", index);
							#endif

							trust = (struct btcnode_protocol_trust_t*) (globals->btserver.clients[index].buffer + sizeof(struct btcnode_protocol_header_t));
						}
						/* message packet */
						else if (he->pkg_type == btcnode_protocol_msg) {
							#ifdef DEBUG_BT_INFO
							printf("bt_client[%d](): msg packet\n", index);
							#endif

							msg = (struct btcnode_protocol_msg_t*) (globals->btserver.clients[index].buffer + sizeof(struct btcnode_protocol_header_t));
							
							ret_val = bt_msg_send(globals, index, globals->btserver.clients[index].buffer, 0);
						}
						/* routing packet */
						else if (he->pkg_type == btcnode_protocol_route) {
							#ifdef DEBUG_BT_INFO
							printf("bt_client[%d](): routing packet\n", index);
							#endif

							route = (struct btcnode_protocol_route_t*) (globals->btserver.clients[index].buffer + sizeof(struct btcnode_protocol_header_t));
							
							/*  TODO
							 *  routing policy check
							 */
							ret_val = bt_node_update(globals, (void*)route, 2, index, NULL);
						}
						/* full routing packet */
						else if (he->pkg_type == btcnode_protocol_route_full) {
							#ifdef DEBUG_BT_INFO
							printf("bt_client[%d](): full routing packet\n", index);
							#endif

							routefull = (struct btcnode_protocol_route_full_t*) (globals->btserver.clients[index].buffer + sizeof(struct btcnode_protocol_header_t));

							/*  TODO
							 *  routing policy check
							 */
							ret_val = bt_node_update(globals, (void*)routefull, 3, index, NULL);
						}
						/* unknown packet */
						else {
							#ifdef DEBUG_BT
							printf("bt_client[%d](): unknown packet!\n", index);
							#endif
							
							globals->btserver.clients[index].buffer_pos = 0;
						}
						
						/* tidy up read buffer */
						memmove(globals->btserver.clients[index].buffer, (globals->btserver.clients[index].buffer + he->length), (globals->btserver.clients[index].buffer_pos - he->length));
						globals->btserver.clients[index].buffer_pos -= he->length;

						/* check buffer again */
						goto check_buffer_again;
					}
					/* packet not complete */
					else {
						/*  change back byte order host->network
						 *  this is ugly - but only happens when receiving real big packets!
						 */
						he->time = htonl(he->time);
						he->length = htons(he->length);
					}
				}
				/* header not complete */
				else {
				}
			}
			/* poll() strange behavior? */
			else {
				#ifdef DEBUG_BT_INFO2
				printf("bt_client[%d](): poll() revents = 0x%x\n", index, fd->revents);
				#endif
			}
		}
		/* socket is corrupted */
		else {
			#ifdef DEBUG_BT
			printf("bt_client[%d](): fd->fd has strange value\n", index);
			printf("bt_client[%d](): closing connection\n", index);
			#endif
		
			goto disconnect;
		}
		
		/* process pending events */
		if (event_get(&globals->btserver.clients[index].events, &event) == 1) {
		
			/* message event */
			if (event->etype == BTCHAT_EVENT_MSG) {
				#ifdef DEBUG_BT_INFO
				printf("bt_client[%d](): message, length = %d\n", index, event->data.msg.msg_length);
				#endif
				
				he = (struct btcnode_protocol_header_t*) event->data.msg.msg;
				/* change byte order network->host */
				he->time = htonl(he->time);
				he->length = htons(he->length);
				
				write_bytes = nwrite(globals->btserver.clients[index].s, event->data.msg.msg, event->data.msg.msg_length);
			}
			/* routing/full_routing event */
			else if (event->etype == BTCHAT_EVENT_ROUTE) {
				#ifdef DEBUG_BT_INFO
				printf("bt_client[%d](): route, length = %d\n", index, event->data.route.data_length);
				routefull = (struct btcnode_protocol_route_full_t*) (event->data.route.data + sizeof(struct btcnode_protocol_header_t));
				ba2str((bdaddr_t*)&routefull->dst, bdstr);
				printf("bt_client[%d](): route %s [%s] hops = %d\n", index, routefull->name, bdstr, routefull->hops);
				#endif
				
				he = (struct btcnode_protocol_header_t*) event->data.route.data;
				/* change byte order network->host */
				he->time = htonl(he->time);
				he->length = htons(he->length);
				
				write_bytes = nwrite(globals->btserver.clients[index].s, event->data.route.data, event->data.route.data_length);
			}
			/* unknown event */
			else {
				event_free(&event);
				
				goto pending_events;
			}

			#ifdef DEBUG_BT_INFO
			printf("bt_client[%d](): write_bytes = %d\n", index, write_bytes);
			#endif
			
			/* increment send statistic counter */
			globals->btserver.clients[index].bytes_send += write_bytes;

			if (write_bytes == 0) {
				/*  TODO
				 *  if a message couldn't be sent because the client disconnected
				 *  try to find a other route - don't just delete it!
				 */
				event_free(&event);
				
				#ifdef DEBUG_BT_INFO
				printf("bt_client[%d](): other side disconnected\n", index);
				#endif
					
				goto disconnect;
			}
			
			/* clean up */
			event_free(&event);
		}
	} /* state == 1 */
	/* unknown state */
	else {
		#ifdef DEBUG_BT
		printf("bt_client[%d](): unknown state = %d\n", index, globals->btserver.clients[index].state);
		#endif
	} /* state == ? */
	
	/* check if any events are pending */
pending_events:
	if (ret_val == 2) return(2);
	if (event_available(&globals->btserver.clients[index].events) == 1) return(2);
	else return(1);
	
	/* disconnect client */
disconnect:
	close(globals->btserver.clients[index].s);
	globals->btserver.clients[index].s = -1;
	globals->btserver.clients[index].selector_index = -1;
	globals->btserver.clients[index].state = 0;

	/* decrement connection counter */
	globals->btserver.clients_num--;
	
	/* update hops for disconnected client */	
	globals->btserver.clients[index].node->routes[index].hops_old = globals->btserver.clients[index].node->routes[index].hops;
	globals->btserver.clients[index].node->routes[index].hops = -1;
	
	if (bt_node_isconnected(globals->btserver.clients[index].node) == 0) {
		if (globals->btserver.clients[index].node->inquiry_flag == 1) {
			globals->btserver.clients[index].node->state = BTCHAT_NODE_STATE_VISIBLE;
		}
		else {
			globals->btserver.clients[index].node->state = BTCHAT_NODE_STATE_OFFLINE;
		}
		
		#ifdef DEBUG_BT_INFO
		printf("bt_client[%d](): disconnected - state = %d inquiry_flag = %d\n", index, globals->btserver.clients[index].node->state, globals->btserver.clients[index].node->inquiry_flag);
		#endif
	}

	/* propagate disconnection of node */
	bt_route_propagate(globals, globals->btserver.clients[index].node, index, 1);
	
	/* update routing tables etc... */
	bt_client_disconnected(globals, index);
	
	event_queue_reset(&globals->btserver.clients[index].events);
				
	return(2);
}

/**
 *  @brief setup a incoming connection
 *
 *  @param globals pointer to globals
 *  @param socket  the connection socket
 *  @param sa      the socket address
 *  @param sa_len  the length of the socket address
 *
 *  @returns -1 on error, 1 when connection was not made and 2 if connection was successful
 */
int bt_client_direct_incoming(struct btchatd_t *globals, int socket, struct sockaddr_rc *sa, int sa_len)
{
	struct btchat_node_t *node;
	struct btcdiscover_protocol_device_t dev;
	int i;
	int index;
	

	#ifdef DEBUG_BT
	assert(globals != NULL);
	#endif

	
	memset(&dev, 0, sizeof(struct btcdiscover_protocol_device_t));
	memcpy(dev.addr, sa->rc_bdaddr.b, 6);
	dev.btchat_channel = -1;
	dev.discovery_method = BTCHAT_DISCOVERY_CONNECT;
	dev.state = BTCHAT_NODE_STATE_VISIBLE;

	/* update/add node information */
	node = NULL;
	bt_node_update(globals, (void*)&dev, 0, 8, &node);
	
	if (node == NULL) {
		#ifdef DEBUG_BT
		printf("bt_client_direct_incoming(): node == NULL, ERROR!\n");
		#endif
		return(-1);
	}

	if (node->trust_route <= BTCHAT_TRUST_LEVEL_ROUTE_DONTCONNECT) {
		close(socket);
		#ifdef DEBUG_BT_INFO
		printf("bt_client_direct_incoming(): trust level is DON'T CONNECT!\n");
		#endif
		return(1);
	}
	
	/* make sure that we don't have a connection to this client all ready */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (node->routes[i].hops == 0) {
			#ifdef DEBUG_BT_INFO
			printf("bt_client_direct_incoming(): all ready connected\n");
			#endif
			return(1);
		}
	}
	
	/* get client index to insert new connection */
	index = -1;
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (globals->btserver.clients[i].s == -1) {
			index = i;
			break;
		}
	}

	#ifdef DEBUG_BT
	if (index == -1) {
		printf("bt_client_direct_incoming(): insert index == -1\n");
		return(-1);
	}
	#endif
	
	/* add routing information */
	node->routes[index].hops = 0;
	node->routes[index].hops_old = -1;
	node->routes[index].last_contact = time(NULL);
	
	/* setup direct connection */
	globals->btserver.clients[index].node = node;
	globals->btserver.clients[index].s = socket;
	globals->btserver.clients[index].sa_len = sa_len;
	memcpy(&globals->btserver.clients[index].sa, sa, sa_len);
	globals->btserver.clients[index].selector_index = -1;
	globals->btserver.clients[index].connection_direction = BTCHAT_CONNECTION_INCOMING;
	globals->btserver.clients[index].state = BTCHAT_NODE_DIRECT_STATE_CONNECTED;
	globals->btserver.clients[index].buffer_pos = 0;
	globals->btserver.clients[index].buffer_size = 0xffff;

	globals->btserver.clients[index].bytes_send = 0;
	globals->btserver.clients[index].bytes_recv = 0;

	/* increment number of connected clients */
	globals->btserver.clients_num++;
	
	/* now set state to online */
	if (node->state < BTCHAT_NODE_STATE_ONLINE) node->state = BTCHAT_NODE_STATE_ONLINE;

	/* first tell other side our name */
	bt_self_publish(globals, index);
		
	/* propagate connection and notify user client */
	bt_route_propagate(globals, node, index, 1);

	/* publish routes */
	bt_route_publish(globals, index);

	return(2);
}

/**
 *  @brief connect to remote node
 *
 *  @param globals pointer to globals
 *  @param node    pointer to node
 *
 *  @returns -1 on error, 1 on failure, 2 on success
 */
int bt_client_direct_outgoing(struct btchatd_t *globals, struct btchat_node_t *node)
{
	int index;
	int i;
	int ret_val = 1;
	#ifdef DEBUG_BT_INFO
	char addr_str[20];
	#endif

	
	#ifdef DEBUG_BT
	assert(globals != NULL);
	assert(node != NULL);
	#endif
	

	#ifdef DEBUG_BT_INFO
	ba2str(&node->addr, addr_str);
	printf("bt_client_direct_outgoing(): connecting: %s [%s]\n", node->name, addr_str);
	#endif
	
	/* find clients index */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (globals->btserver.clients[i].s == -1) {
			index = i;
			break;
		}
	}
		
	/* init bt client */
	if ((globals->btserver.clients[index].s = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		#ifdef DEBUG_BT
		perror("bt_client_direct_outgoing(): socket()");
		#endif
		ret_val = -1;
		globals->btserver.clients[index].s = -1;
		goto out;
	}
		
	/* bind to local address */
	globals->btserver.clients[index].sa.rc_family = AF_BLUETOOTH;
	globals->btserver.clients[index].sa.rc_channel = 0;
	globals->btserver.clients[index].sa.rc_bdaddr = globals->self.addr;
		
	if (bind(globals->btserver.clients[index].s, (struct sockaddr*) &globals->btserver.clients[index].sa, sizeof(globals->btserver.clients[index].sa)) < 0) {
		#ifdef DEBUG_BT
		perror("bt_client_direct_outgoing(): bind()");
		#endif
		ret_val = -1;
		globals->btserver.clients[index].s = -1;
		goto out;
	}
		
	/* connect to remote device */
	globals->btserver.clients[index].sa.rc_family = AF_BLUETOOTH;
	globals->btserver.clients[index].sa.rc_bdaddr = node->addr;
	globals->btserver.clients[index].sa.rc_channel = node->btchat_channel;
		
	if (connect(globals->btserver.clients[index].s, (struct sockaddr*) &globals->btserver.clients[index].sa, sizeof(globals->btserver.clients[index].sa)) < 0) {
		#ifdef DEBUG
		perror("bt_client_direct_outgoing(): connect()");
		#endif
		ret_val = -1;
		globals->btserver.clients[index].s = -1;
		goto out;
	}
		
	/* if connect is successful setup client data */
	globals->btserver.clients[index].selector_index = -1;
	globals->btserver.clients[index].connection_direction = BTCHAT_CONNECTION_OUTGOING;
	globals->btserver.clients[index].node = node;
	globals->btserver.clients[index].buffer_pos = 0;
	globals->btserver.clients[index].buffer_size = 0xffff;
	globals->btserver.clients[index].state = BTCHAT_NODE_DIRECT_STATE_CONNECTED;

	globals->btserver.clients[index].bytes_send = 0;
	globals->btserver.clients[index].bytes_recv = 0;

	/* update node data */
	node->routes[index].last_contact = time(NULL);
	node->routes[index].hops = 0;
	node->routes[index].hops_old = -1;

	if (node->state < BTCHAT_NODE_STATE_ONLINE) node->state = BTCHAT_NODE_STATE_ONLINE;

	/* increment number of connected clients */
	globals->btserver.clients_num++;
		
	#ifdef DEBUG_BT_INFO
	printf("bt_client_direct_outgoing(): connected\n");
	#endif

	/* indicate event creation */
	ret_val = 2;

	/* first tell other side our name */
	bt_self_publish(globals, index);

	/* propagate new connection - to other nodes */
	bt_route_propagate(globals, node, index, 1);
	
	/* publish routes */
	bt_route_publish(globals, index);

/* connection failed */
out:

	return(ret_val);
}

/*
 *  $Log$
 */
