/*
 *  $Id$
 */
 
/**
 *  @file bt_discover.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "node_protocol.h"
#include "event.h"
#include "util.h"
#include "discover_protocol.h"
#include "bt_discover.h"
#include "sdp.h"
#include "bt_net.h"

/* declared in main.c */
extern int hci_dev;

/**
 *  @brief compare two discover_nodes for searching the tree
 *
 *  @param d1 pointer to left node
 *  @param d2 pointer to right node
 *
 *  @returns 0 if equal, 1 if d1 > d2 and -1 if d2 > d1
 */
static int btd_tree_cmp(const void *d1, const void *d2)
{
	struct btchat_discover_node_t *a1 = (struct btchat_discover_node_t*) d1;
	struct btchat_discover_node_t *a2 = (struct btchat_discover_node_t*) d2;
	
	#ifdef DEBUG_BT_DISCOVER
	assert(a1 != NULL);
	assert(a2 != NULL);
	#endif
	
	return(memcmp(a1->addr.b, a2->addr.b, 6));
}

/**
 *  @brief send devinfo to discover client if given node has changed status
 *
 *  @param node pointer to the node to check
 *  @param data pointer to bt_discover_server globals
 *
 *  @returns 1
 */
static int btd_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_discover_node_t *n = NULL;
	struct btchat_btdiscover_server_t *g = NULL;
	unsigned char out_buffer[1024];
	int out_buffer_size = 1024;
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_device_t *dev;
	int write_bytes = 0;
	
	
	#ifdef DEBUG_BT_DISCOVER
	assert(node != NULL);
	assert(data != NULL);
	#endif


	n = (struct btchat_discover_node_t*) node->data;
	g = (struct btchat_btdiscover_server_t*) data;

	/* detect state change! */
	if (n->last < g->current_time && n->state != BTCHAT_NODE_STATE_OFFLINE) {
		n->state = BTCHAT_NODE_STATE_OFFLINE;
	}
	else if (n->updated == 1) {
	}
	else if (n->state == BTCHAT_NODE_STATE_OFFLINE && n->last == g->current_time) {
		n->state = BTCHAT_NODE_STATE_VISIBLE;
	}
	else {
		#ifdef DEBUG_BT_DISCOVER_INFO
		printf("btd_tree_walk(): nothing happened name = %s\n", n->name);
		#endif
		return(1);
	}

	memset(out_buffer, 0, out_buffer_size);
	he = (struct btcdiscover_protocol_header_t*) out_buffer;
	he->version = 0x01;
	he->pkg_type = btcdiscover_protocol_device;
	he->time = g->current_time;
	he->length = (sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_device_t));
						
	dev = (struct btcdiscover_protocol_device_t*) (out_buffer + sizeof(struct btcdiscover_protocol_header_t));
	memcpy(dev->addr, n->addr.b, 6);
								
	memcpy(dev->name, n->name, 250);
		
	dev->device_class = n->device_class;
	dev->discovery_method = n->discovery_method;
								
	dev->btchat_channel = n->btchat_channel;
			
	/* reset updated flag */
	n->updated = 0;
	dev->state = n->state;
		
	#ifdef DEBUG_BT_DISCOVER_INFO
	printf("btd_tree_walk(): name = %s\n", dev->name);
	#endif
		
	write_bytes = nwrite(g->s, out_buffer, sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_device_t));
	if (write_bytes <= 0) {
		#ifdef DEBUG_BT_DISCOVER
		printf("btd_tree_walk(): write_bytes <= 0, ERROR!\n");
		#endif
		
		/* break tree walk on send error */
		return(2);
	}
	else {
		#ifdef DEBUG_BT_DISCOVER_INFO
		printf("btd_tree_walk(): write_bytes = %d\n", write_bytes);
		#endif
	}

	return(1);
}

/**
 *  @brief query a device to get the name and the btchat rfcomm channel
 *
 *  @param hci   the open hci interface
 *  @param local the local bd_addr to use when connecting device (for sdp)
 *  @param node  pointer to the node to query
 *  @param init  initial query (1=yes, 0=no)
 *
 *  @returns 1
 */
static int bt_discover_query_dev(int hci, bdaddr_t *local, struct btchat_discover_node_t *node, int init)
{
	char tmp_name[250];
	int res;
	int tmp_channel = 0;
	
	
	#ifdef DEBUG_BT_DISCOVER
	assert(local != NULL);
	assert(node != NULL);
	#endif
	

	/* get name */
	memset(tmp_name, 0, sizeof(tmp_name));
	res = hci_read_remote_name(hci, &node->addr, sizeof(tmp_name), tmp_name, 10000);
	if (res < 0) {
		#ifdef DEBUG_BT_DISCOVER
		perror("bt_discover_query_dev(): hci_read_remote_name()");
		#endif
		
		if (init == 1) ba2str(&node->addr, node->name);
	}
	else {
		memcpy(node->name, tmp_name, sizeof(node->name));
	}
	#ifdef DEBUG_BT_DISCOVER_INFO
	ba2str(&node->addr, tmp_name);
	printf("bt_discover_query_dev(): %s [%s]\n", node->name, tmp_name);
	#endif
								
	/* get btChat channel */
	res = bt_sdp_search(local, &node->addr, &tmp_channel);
	//res = bt_sdp_search(BDADDR_ANY, &node->addr, &tmp_channel);
	if (res == 1) {
		node->btchat_channel = tmp_channel;
		#ifdef DEBUG_BT_DISCOVER_INFO
		printf("bt_discover_query_dev(): btChat channel = %d\n", node->btchat_channel);
		#endif
	}
	else {
		if (init == 1) node->btchat_channel = -1;
	}
	
	return(1);
}

/**
 *  @brief discover other bluetooth devices
 *
 *  This function runs as a separate process (forked of by main)
 *
 *  @param socket the communication socket
 *
 *  @returns allays returns 1
 */
int bt_discover_server(int socket)
{
	struct btchat_btdiscover_server_t *dg = NULL;
	struct pollfd selectors[1];
	int selectors_ready;
	int read_bytes;
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_discover_t *dis;
	struct btcdiscover_protocol_devinfo_t *dinfo;
	inquiry_info *info = NULL;
	int flags = 0;
	int num_rsp = 0;
	int i;
	#ifdef DEBUG_BT_DISCOVER_INFO
	char bastr[20];
	#endif
	int res;
	struct btchat_discover_node_t node;
	struct btchat_discover_node_t *new_node;
	struct tree_node_t *result_node;
	int hci_dd;
	bdaddr_t local_addr;
		

	/* allocate globals */
	dg = (struct btchat_btdiscover_server_t*)	malloc(sizeof(struct btchat_btdiscover_server_t));
	
	#ifdef DEBUG_BT_DISCOVER
	assert(dg != NULL);
	#endif
	
	dg->s = socket;
	
	dg->buffer_pos = 0;
	dg->buffer_size = 1024;
	
	/* copy hci_dev id */
	dg->bt_device = hci_dev;
	/* get local bd address */
	hci_devba(dg->bt_device, &local_addr);
	
	/* init b-tree */
	dg->btdevices.splay = 1;
	dg->btdevices.root = NULL;
	dg->btdevices.num = 0;
	dg->btdevices.compare_function = btd_tree_cmp;
	dg->btdevices.walk_function = btd_tree_walk;

	/* bt_discover_server() main loop */	
	for (;;) {
		memset(selectors, 0, sizeof(selectors));
		selectors[0].fd = dg->s;
		selectors[0].events = POLLIN | POLLPRI;
		
		/* wait until a command is received from the client */
		selectors_ready = poll(selectors, 1, -1);
		
		/* check poll(..) results */
		if (selectors_ready != 1) {
			continue;
		}
		if ((selectors[0].revents & POLLIN) != POLLIN && (selectors[0].revents & POLLPRI) != POLLPRI) {
			continue;
		}
		
		#ifdef DEBUG_BT_DISCOVER_INFO
		printf("bt_discover_server(): socket is ready to read!\n");
		#endif
			
		read_bytes = cread(dg->s, (dg->buffer + dg->buffer_pos), (dg->buffer_size - dg->buffer_pos));
			
		#ifdef DEBUG_BT_DISCOVER_INFO
		printf("bt_discover_server(): read_bytes = %d\n", read_bytes);
		#endif
			
		if (read_bytes == 0) {
			#ifdef DEBUG_BT_DISCOVER
			printf("bt_discover_server(): other end closed communication socket, this normally means exit, terminating\n");
			#endif
			break;
		}			
		
		/* advance buffer position */
		dg->buffer_pos += read_bytes;
			
		/* check for protocol header size */
		if (dg->buffer_pos >= sizeof(struct btcdiscover_protocol_header_t)) {
		
			he = (struct btcdiscover_protocol_header_t*) dg->buffer;
				
			/* discover devices */
			if (he->pkg_type == btcdiscover_protocol_discover || he->pkg_type == btcdiscover_protocol_btchatdiscover) {
				if ((dg->buffer_size - sizeof(struct btcdiscover_protocol_header_t)) >= sizeof(struct btcdiscover_protocol_discover_t)) {
					
					dis = (struct btcdiscover_protocol_discover_t*) (dg->buffer + sizeof(struct btcdiscover_protocol_header_t));
						
					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_server(): doing inquiry for %d seconds\n", dis->inquiry_time);
					#endif
						
					/* do bluetooth inquiry */
				inquiry:
					num_rsp = hci_inquiry(dg->bt_device, dis->inquiry_time, 255, NULL, &info, flags);
					
					if (num_rsp < 0) {
						#ifdef DEBUG_BT_DISCOVER
						printf("bt_discover_server(): hci_inquiry() failed ... try to reinitialize hci device .. sleeping 10sec\n");
						sleep(10);
						#endif

						goto inquiry;
					}
					
					/* tidy up read buffer - here */
					memmove(dg->buffer, (dg->buffer + he->length), (dg->buffer_pos - he->length));
					dg->buffer_pos -= he->length;
						
					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_server(): buffer_pos = %d\n", dg->buffer_pos);
					#endif

					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_server(): num devices = %d\n", num_rsp);
					#endif
					
					/* get current time */
					time(&dg->current_time);
					
					/* open hci device for name requests */
					hci_dd = hci_open_dev(dg->bt_device);
					if (hci_dd < 0) {
						#ifdef DEBUG_BT_DISCOVER
						perror("bt_discover_server(): hci_open_dev()");
						#endif
					}
					
					/* process inquiry results */
					for (i = 0; i < num_rsp; i++) {
						memset(&node, 0, sizeof(struct btchat_discover_node_t));
						memcpy(node.addr.b, &(info+i)->bdaddr.b, 6);
							
						#ifdef DEBUG_BT_DISCOVER_INFO
						ba2str(&node.addr, bastr);
						printf("bt_discover_server(): dev = %d addr = %s\n", i, bastr);
						#endif
							
						result_node = NULL;
						res = tree_access_insert(&dg->btdevices, (void*) &node, &result_node);
						
						switch (res) {
						case 1:	/* known node */
							new_node = (struct btchat_discover_node_t*) result_node->data;
							
							#ifdef DEBUG_BT_DISCOVER_INFO
							printf("bt_discover_server(): known node state = %d\n", new_node->state);
							#endif
							
							if (new_node->state == BTCHAT_NODE_STATE_OFFLINE) {
								bt_discover_query_dev(hci_dd, &local_addr, new_node, 0);
							}
							
							new_node->last = dg->current_time;
							break;
						case 0:	/* new node */
							#ifdef DEBUG_BT_DISCOVER_INFO
							ba2str(&node.addr, bastr);
							printf("bt_discover_server(): unknown node dev = %d [%s]\n", i, bastr);
							#endif
					
							result_node->data = malloc(sizeof(struct btchat_discover_node_t));
							new_node = (struct btchat_discover_node_t*) result_node->data;
							memset(new_node, 0, sizeof(struct btchat_discover_node_t));
							
							bacpy(&new_node->addr, &node.addr);
							new_node->discovery_method = BTCHAT_DISCOVERY_INQUIRY;
							
							bt_discover_query_dev(hci_dd, &local_addr, new_node, 1);
					
							new_node->last = dg->current_time;
							new_node->updated = 1;
							new_node->state = BTCHAT_NODE_STATE_VISIBLE;
							
							memcpy(((uint8_t*)&new_node->device_class), (info+i)->dev_class, 3);
							break;
						default:
							#ifdef DEBUG_BT_DISCOVER
							printf("bt_discover_server(): res unknown - ERROR\n");
							#endif
							break;
						}
					} /* for (i < num_rsp) */
					
					/* close hci device */
					if (hci_dd > 0) {
						close(hci_dd);
					}
					
					/* free inquiry information */
					if (info != NULL) {
						free(info);
						info = NULL;
					}

					/* check all nodes for change */
					if (tree_walk(&dg->btdevices, dg) == 2) {
						#ifdef DEBUG_BT_DISCOVER
						printf("bt_discover_server(): tree_walk reports: lost connection, terminating!\n");
						break;
						#endif
					}
				}
				else {
					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_server(): body data to small - waiting for more\n");
					#endif
					continue;
				}
			}
			/* get device information */
			else if (he->pkg_type == btcdiscover_protocol_devinfo) {
				#ifdef DEBUG_BT_DISCOVER
				printf("bt_discover_server(): protocol_devinfo\n");
				#endif
				
				if ((dg->buffer_size - sizeof(struct btcdiscover_protocol_header_t)) >= sizeof(struct btcdiscover_protocol_devinfo_t)) {
		
					dinfo = (struct btcdiscover_protocol_devinfo_t*) (dg->buffer + sizeof(struct btcdiscover_protocol_header_t));
					
					memset(&node, 0, sizeof(struct btchat_discover_node_t));
					memcpy(node.addr.b, &dinfo->addr, 6);
							
					#ifdef DEBUG_BT_DISCOVER_INFO
					ba2str(&node.addr, bastr);
					printf("bt_discover_server(): devinfo [%s]\n", bastr);
					#endif
					
					/* open hci device for name requests */
					hci_dd = hci_open_dev(dg->bt_device);
					if (hci_dd < 0) {
						#ifdef DEBUG_BT_DISCOVER
						perror("bt_discover_server(): hci_open_dev()");
						#endif
					}
					
					result_node = NULL;
					res = tree_access_insert(&dg->btdevices, (void*) &node, &result_node);
					
					switch (res) {
					case 1:	/* known node */
						new_node = (struct btchat_discover_node_t*) result_node->data;

						bt_discover_query_dev(hci_dd, &local_addr, new_node, 0);
				
						new_node->last = dg->current_time;
						break;
					case 0:	/* new node */
						#ifdef DEBUG_BT_DISCOVER_INFO
						ba2str(&node.addr, bastr);
						printf("bt_discover_server(): unknown node [%s]\n", bastr);
						#endif
					
						result_node->data = malloc(sizeof(struct btchat_discover_node_t));
						new_node = (struct btchat_discover_node_t*) result_node->data;
						memset(new_node, 0, sizeof(struct btchat_discover_node_t));
							
						bacpy(&new_node->addr, &node.addr);
						new_node->discovery_method = BTCHAT_DISCOVERY_INQUIRY;

						bt_discover_query_dev(hci_dd, &local_addr, new_node, 1);
							
						new_node->last = dg->current_time;
						new_node->updated = 1;
						new_node->state = BTCHAT_NODE_STATE_VISIBLE;	
						break;
					default:
						#ifdef DEBUG_BT_DISCOVER
						printf("bt_discover_server(): res unknown - ERROR\n");
						#endif
						break;
					}
					
					/* close hci device */
					if (hci_dd > 0) {
						close(hci_dd);
					}
					
					/* directly call tree walk to just update this node */
					btd_tree_walk(result_node, dg);
				}
				else {
					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_server(): body data to small - waiting for more\n");
					#endif
					continue;
				}
			}
			else if (he->pkg_type == btcdiscover_protocol_null) {
				#ifdef DEBUG_BT_DISCOVER
				printf("bt_discover_server(): protocol_null\n");
				#endif
				continue;
			}
			else {
				#ifdef DEBUG_BT_DISCOVER
				printf("bt_discover_server(): protocol unknown\n");
				#endif
				continue;
			}
		}
		else {
			#ifdef DEBUG_BT_DISCOVER_INFO
			printf("bt_discover_server(): header data to small - waiting for more\n");
			#endif
			continue;
		}
	} /* for (;;) */
	
	return(1);
}

/**
 *  @brief talk to bluetooth device discovery server
 *
 *  @param globals pointer to globals
 *  @param fd      pointer to poll events
 *
 *  @returns -1 on error 1 on success and 2 on pending or generated events
 */
int bt_discover_client(struct btchatd_t *globals, struct pollfd *fd)
{
	struct btcdiscover_protocol_header_t *he;
	struct btcdiscover_protocol_device_t *dev;
	struct btcdiscover_protocol_discover_t *dis;
	struct btcdiscover_protocol_devinfo_t *dinfo;
	int read_bytes = 0;
	struct btchat_event_t *event = NULL;
	unsigned char out_buffer[1024];
	int i;
	int res;
	int num_devices;
	int ret_val = 1;
	#ifdef DEBUG_BT_DISCOVER_INFO
	char bastr[20];
	#endif
	
	
	#ifdef DEBUG_BT_DISCOVER
	assert(globals != NULL);
	#endif
	
	
	if (fd == NULL) {
		#ifdef DEBUG_BT_DISCOVER
		printf("bt_discover_client(): fd == NULL\n");
		#endif
		return(-1);
	}
		
	if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
	
		read_bytes = cread(globals->btdiscover.s, globals->btdiscover.buffer + globals->btdiscover.buffer_pos, globals->btdiscover.buffer_size - globals->btdiscover.buffer_pos);
		if (read_bytes == 0) {
			#ifdef DEBUG_BT_DISCOVER
			printf("bt_discover_client(): read_bytes = 0, looks like the bt_discover_server() closed the socket, terminate\n");
			#endif
			return(1);
		}
		globals->btdiscover.buffer_pos += read_bytes;

	check_buffer_again:
		if (globals->btdiscover.buffer_pos >= sizeof(struct btcdiscover_protocol_header_t)) {
			he = (struct btcdiscover_protocol_header_t*) globals->btdiscover.buffer;
				
			#ifdef DEBUG_BT_DISCOVER_INFO
			printf("bt_discover_client(): he->length = %d, buffer_pos = %d\n", he->length, globals->btdiscover.buffer_pos);
			#endif
				
			if (globals->btdiscover.buffer_pos >= he->length) {
			
				/* device info packet */
				if (he->pkg_type == btcdiscover_protocol_device) {
					num_devices = (he->length - sizeof(struct btcdiscover_protocol_header_t)) / sizeof(struct btcdiscover_protocol_device_t);
						
					#ifdef DEBUG_BT_DISCOVER_INFO
					printf("bt_discover_client(): %d devices\n", num_devices);
					#endif
						
					for (i = 0; i < num_devices; i++) {
						dev = (struct btcdiscover_protocol_device_t*) (globals->btdiscover.buffer + sizeof(struct btcdiscover_protocol_header_t) + (i * sizeof(struct btcdiscover_protocol_device_t)));
						
						/* update each node */
						ret_val = bt_node_update(globals, (void*)dev, 0, 7, NULL);
												
					} /* for (i < num_devices) */
					
					/* tidy up read buffer */
					memmove(globals->btdiscover.buffer, (globals->btdiscover.buffer + he->length), (globals->btdiscover.buffer_pos - he->length));
					globals->btdiscover.buffer_pos -= he->length;

					/* check buffer again */
					goto check_buffer_again;
				}
				else {
					/* unknown pkg_type - rewind buffer */
					globals->btdiscover.buffer_pos = 0;
				}
			}
			else {
				/* package not complete */
			}
		}
		else {
		}
	} /* fd->revents & POLLIN */
	
	/* process pending events */
	if (event_get(&globals->btdiscover.events, &event) == 1) {

		memset(out_buffer, 0, 1024);
		he = (struct btcdiscover_protocol_header_t*) out_buffer;
	
		/* discover event */
		if (event->etype == BTCHAT_EVENT_DISCOVER) {
			dis = (struct btcdiscover_protocol_discover_t*) (out_buffer + sizeof(struct btcdiscover_protocol_header_t));
			
			he->version = 0x01;
			he->pkg_type = btcdiscover_protocol_discover;
			he->length = sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_discover_t);
			he->time = time(NULL);
			dis->inquiry_time = globals->btdiscover.inquiry_duration;
			
			#ifdef DEBUG_BT_DISCOVER_INFO
			printf("bt_discover_client(): inquiry_duration = %d\n", dis->inquiry_time);
			#endif

			res = nwrite(globals->btdiscover.s, out_buffer, he->length);
			
			#ifdef DEBUG_BT_DISCOVER_INFO
			printf("bt_discover_client: wrote = %d\n", res);
			#endif
			
			if (res != he->length) {
				#ifdef DEBUG_BT_DISCOVER
				printf("bt_discover_client(): nwrite failed\n");
				#endif
			}
		}
		/* get device info event */
		else if (event->etype == BTCHAT_EVENT_DEVINFO) {
			dinfo = (struct btcdiscover_protocol_devinfo_t*) (out_buffer + sizeof(struct btcdiscover_protocol_header_t));
			
			he->version = 0x01;
			he->pkg_type = btcdiscover_protocol_devinfo;
			he->length = sizeof(struct btcdiscover_protocol_header_t) + sizeof(struct btcdiscover_protocol_devinfo_t);
			he->time = time(NULL);
			
			memcpy(dinfo->addr, event->data.node.node_ptr->addr.b, sizeof(dinfo->addr));
			
			#ifdef DEBUG_BT_DISCOVER_INFO
			ba2str(&event->data.node.node_ptr->addr, bastr);
			printf("bt_discover_client(): get information about [%s]\n", bastr);
			#endif

			res = nwrite(globals->btdiscover.s, out_buffer, he->length);
			
			#ifdef DEBUG_BT_DISCOVER_INFO
			printf("bt_discover_client: wrote = %d\n", res);
			#endif
			
			if (res != he->length) {
				#ifdef DEBUG_BT_DISCOVER
				printf("bt_discover_client(): nwrite failed\n");
				#endif
			}
		}
		else {
			#ifdef DEBUG_BT_DISCOVER
			printf("bt_discover_client(): unknown event\n");
			#endif
		}
		
		free(event);
		event = NULL;
	}

	return(ret_val);
}

/*
 *  $Log$
 */
