/*
 *  $Id$
 */
 
/**
 *  @file bt_net.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "util.h"
#include "bt_discover.h"
#include "node_protocol.h"
#include "discover_protocol.h"
#include "client_protocol.h"
#include "event.h"
#include "bt_net.h"

/**
 *  @brief select a node to be connected (walk_tree callback)
 *
 *  @param node pointer to the current node
 *  @param data pointer to connect_data struct
 *
 *  @param returns 1 when node is not suitable or 2 when node should be connected
 */
static int bt_client_connect_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_node_t *n;
	struct btchat_connect_data_out_t *g;
	int i;
	int hops_num = 0xffff;
	int hops_index = -1;
	

	#ifdef DEBUG
	assert(node != NULL);
	assert(data != NULL);
	#endif


	n = (struct btchat_node_t*) node->data;
	g = (struct btchat_connect_data_out_t*) data;

	/* make sure the node isn't offline */
	if (n->state == BTCHAT_NODE_STATE_OFFLINE) {
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_client_connect_tree_walk(): %s state = %d offline\n", n->name[0] != 0 ? n->name : n->alias, n->state);
		#endif
		return(1);
	}

	/* make sure the node is running btChat */
	if (n->btchat_channel == -1) {
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_client_connect_tree_walk(): %s state = %d no btchat channel\n", n->name[0] != 0 ? n->name : n->alias, n->state);
		#endif
		return(1);
	}
		
	/* check last contact */
	if (g->time_index < n->contact_last && g->n != NULL) {
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_client_connect_tree_walk(): %s state = %d last contact\n", n->name[0] != 0 ? n->name : n->alias, n->state);
		#endif
		return(1);
	}

	/*  check routes to make sure that no connections to this route exist
	 *  also find shortest route to node for later comparison with current
	 *  selected node/route
	 */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (n->routes[i].hops == -1) continue;

		/*  make sure the node is not connected all ready 
		 *  or if the distance is very short (1 hop)
		 */
		if (n->routes[i].hops >= 0 && n->routes[i].hops < g->g->btserver.connect_min_distance) {
			#ifdef DEBUG_BTNET_INFO2
			printf("bt_client_connect_tree_walk(): hops = %d\n", n->routes[i].hops);
			#endif
			return(1);
		}
		else {
			/* find shortest route */
			if (hops_num > n->routes[i].hops) {
				hops_num = n->routes[i].hops;
				hops_index = i;
			}
		}
	}
	
	if (hops_index != -1) {
		/* don't remember node with a smaller distance */
		if (g->hops > hops_num) {
			#ifdef DEBUG_BTNET_INFO2
			printf("bt_client_connect_tree_walk(): %s state = %d hops_num = %d\n", n->name[0] != 0 ? n->name : n->alias, n->state, hops_num);
			#endif
			return(1);
		}
	}

	/* save current node */
	g->n = n;
	/* save last contact time */
	g->time_index = n->contact_last;
	
	/* node has no connection/routing information - stop searching and connect */
	if (hops_index == -1) {
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_client_connect_tree_walk(): ret_val = 2\n");
		#endif
		return(2);
	}
	else {
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_client_connect_tree_walk(): ret_val = 1\n");
		#endif
		return(1);
	}
}

/**
 *  @brief connect to other nodes
 *
 *  @param globals pointer to globals
 *
 *  @returns 1
 */
int bt_client_connect(struct btchatd_t *globals)
{
	void *old_walk;
	int res;
	int ret_val = 1;
	struct btchat_connect_data_out_t cd;
	time_t cur_time;

	
	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	#endif


	/* check connection/node limit */
	if (globals->btserver.clients_num >= globals->btserver.clients_max) return(0);

	/* try making outgoing connections every X seconds */
	cur_time = time(NULL);
	if (cur_time < (globals->btserver.connect_last + globals->btserver.connect_delay)) return(0);

	#ifdef DEBUG_BTNET_INFO2
	printf("bt_client_connect(): searching for node\n");
	#endif
	
	/* save old walk function */
	old_walk = globals->local_nodes.walk_function;
	
	globals->local_nodes.walk_function = bt_client_connect_tree_walk;

	cd.g = globals;
	cd.n = NULL;
	cd.hops = 0xffff;
	cd.index = -1;
	cd.time_index = 0x0fffffff;

	res = tree_walk(&globals->local_nodes, (void*)&cd);
	
	globals->local_nodes.walk_function = old_walk;
	
	/* connect node */
	if (cd.n != NULL) {
		cd.n->contact_last = cur_time;
		ret_val = bt_client_direct_outgoing(globals, cd.n);
	}

	/* mark last try */
	globals->btserver.connect_last = cur_time;
	
	return(ret_val);
}

/**
 *  @brief check if a given node is connected
 *
 *  @param n pointer to node
 *
 *  @returns 1 if connected or 0 if not connected
 */
int bt_node_isconnected(struct btchat_node_t *n)
{
	int i;
	
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (n->routes[i].hops != -1) return(1);
	}
	
	return(0);
}

/**
 *  @brief check if a given node is the local node (globals->self)
 *
 *  @param globals pointer to globals
 *  @param n       pointer to node
 *
 *  @returns 1 if node is local/self or 0 if not
 */
static int bt_node_isself(struct btchatd_t *globals, struct btchat_node_t *n)
{
	return(((memcmp(globals->self.addr.b, n->addr.b, sizeof(n->addr.b)) == 0) ? 1 : 0));
}

/**
 *  @brief check if a given node is broken
 *
 *  @param globals pointer to globals
 *  @param n       pointer to node
 *
 *  @returns 1 if node is broken else 0
 */
static int bt_node_isbroken(struct btchatd_t *globals, struct btchat_node_t *n)
{
	const unsigned char broken[6] = {0x00,0x00,0x00,0x00,0x00,0x00};
	return(((memcmp(broken, n->addr.b, sizeof(n->addr.b)) == 0) ? 1 : 0));
}

/**
 *  @brief check if a given node is broadcast
 *
 *  @param globals pointer to globals
 *  @param n       pointer to node
 *
 *  @returns 1 if node is broadcast else 0
 */
static int bt_node_isbcast(struct btchatd_t *globals, struct btchat_node_t *n)
{
	const unsigned char bcast[6] = {0xff,0xff,0xff,0xff,0xff,0xff};
	return(((memcmp(bcast, n->addr.b, sizeof(n->addr.b)) == 0) ? 1 : 0));
}

/**
 *  @brief add/update node
 *
 *  @param globals pointer to globals
 *  @param n       pointer to node data
 *  @param b_type  type of node data
 *  @param src     source of node data
 *  @param out     pointer to pointer to pass back the node data (if NULL no data will be passed back)
 *
 *  @returns 1 if no event was generated and 2 if event was generated, if src == 8 no event will be generated
 */
int bt_node_update(struct btchatd_t *globals, void *n, int n_type, int src, struct btchat_node_t **out)
{
	struct btcdiscover_protocol_device_t *dev;
	struct btchat_node_t *node;
	struct btcnode_protocol_route_t *route;
	struct btcnode_protocol_route_full_t *fullroute;
	struct btchat_node_t *new_node;
	struct btchat_node_t work_node;
	struct tree_node_t *result_node;
	struct btchat_event_t *e;
	int i;
	int ret_val = 1;
	int res;
	int selfpub = 0;
	#ifdef DEBUG_BTNET
	char addr_str[20];
	#endif
	

	/* build current work node struct */
	memset(&work_node, 0, sizeof(struct btchat_node_t));
	
	/* device discovery or incoming connect */
	if (n_type == 0) {
		dev = (struct btcdiscover_protocol_device_t*) n;
	
		#ifdef DEBUG_BTNET
		if (src == 7) printf("bt_node_update(): n_type == discovery\n");
		if (src == 8) printf("bt_node_update(): n_type == incoming connect\n");
		#endif
	
		memcpy(work_node.addr.b, dev->addr, sizeof(work_node.addr.b));

		work_node.discovery_method = dev->discovery_method;
		work_node.btchat_channel = dev->btchat_channel;
							
		work_node.trust_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trust_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
		
		work_node.trusted_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trusted_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
		
		memcpy(work_node.name, dev->name, sizeof(work_node.name));
		
		work_node.device_class = dev->device_class;
		work_node.state = dev->state;
		
		if (src == 7) work_node.inquiry_flag = 1;
	}
	/* some node update */
	else if (n_type == 1) {
		node = (struct btchat_node_t*) n;
		memcpy(&work_node, node, sizeof(struct btchat_node_t));
		
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): n_type == some node update\n");
		#endif	
	}
	/* routing */
	else if (n_type == 2) {
		route = (struct btcnode_protocol_route_t*) n;
		
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): n_type == routing\n");
		#endif
		
		memcpy(work_node.addr.b, route->dst, sizeof(work_node.addr.b));
		work_node.state = route->state;
		work_node.btchat_channel = -1;
		work_node.discovery_method = BTCHAT_DISCOVERY_NET;
		
		work_node.trust_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trust_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
				
		work_node.trusted_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trusted_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
	}
	/* full routing */
	else if (n_type == 3) {
		fullroute = (struct btcnode_protocol_route_full_t*) n;

		#ifdef DEBUG_BTNET
		printf("bt_node_update(): n_type == full routing\n");
		#endif

		memcpy(work_node.addr.b, fullroute->dst, sizeof(work_node.addr.b));
		memcpy(work_node.alias, fullroute->name, sizeof(work_node.name));

		work_node.state = fullroute->state;
		work_node.btchat_channel = -1;		
		work_node.discovery_method = BTCHAT_DISCOVERY_NET;

		work_node.trust_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trust_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
		
		work_node.trusted_route = BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL;
		work_node.trusted_general = BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL;
	}
	/* unknown */
	else {
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): n_type == unknown\n");
		#endif
	}

	#ifdef DEBUG_BTNET
	ba2str(&work_node.addr, addr_str);
	if (strlen(work_node.alias) > 0) printf("bt_node_update(): %s [%s]\n", work_node.alias, addr_str);
	else printf("bt_node_update(): %s [%s]\n", work_node.name, addr_str);
	#endif

	/* ignore information about our self's */
	if (bt_node_isself(globals, &work_node) == 1) {
		return(1);
	}
	/* ignore broken addresses (00:00:00...) */
	if (bt_node_isbroken(globals, &work_node) == 1) {
		return(1);
	}
	/* ignore broadcast addresses (ff:ff:ff...) */
	if (bt_node_isbcast(globals, &work_node) == 1) {
		return(1);
	}

	/* reset hops */	
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		work_node.routes[i].hops = -1;
		work_node.routes[i].hops_old = -1;
	}

	/* search for device/node */
	result_node = NULL;
	res = tree_access_insert(&globals->nodes, &work_node, &result_node);
	
	/* new node */
	if (res == 0) {
	
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): new node\n");
		#endif
	
		result_node->data = (struct btchat_node_t*) malloc(sizeof(struct btchat_node_t));
		new_node = (struct btchat_node_t*) result_node->data;
		memcpy(new_node, &work_node, sizeof(struct btchat_node_t));
		ret_val = 2;
	}
	/* known node */
	else if (res == 1) {
		new_node = (struct btchat_node_t*) result_node->data;
		
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): known node\n");
		#endif

		/* node has publish information about itself */
		if (src >= 0 && src <= 6) {
			if (new_node->routes[src].hops == 0) {
				work_node.state = new_node->state;
				selfpub = 1;
			}
		}

		/* device found thru incoming connection */
		if	(src == 8) {
			/* only change state if state is "not online" */
			if (new_node->state < BTCHAT_NODE_STATE_ONLINE) new_node->state = work_node.state;
		}
		else {		
			/* state change happened */
			if (new_node->state != work_node.state) {
			
				/* node is connected to the net in some way */
				if (bt_node_isconnected(new_node) == 1) {
				
					if (work_node.state >= BTCHAT_NODE_STATE_ONLINE) {
						new_node->state = work_node.state;
						ret_val = 3;
					}
					else {
						#ifdef DEBUG_BTNET
						printf("bt_node_update(): online - state change: old=%d new=%d\n", new_node->state, work_node.state);
						#endif
					}
				}
				/* node is not connected */
				else {
					new_node->state = work_node.state;
					ret_val = 3;
				}
			}
			
			/* don't update the channel when data comes from routing */
			if (n_type != 2 && n_type != 3) new_node->btchat_channel = work_node.btchat_channel;
			/* only overwrite with better (non-zero) information */
			if (strlen(work_node.name) > 0) memcpy(new_node->name, work_node.name, sizeof(new_node->name));
			if (strlen(work_node.alias) > 0) memcpy(new_node->alias, work_node.alias, sizeof(new_node->alias));
			if (work_node.device_class != 0) new_node->device_class = work_node.device_class;
		}
	}
	/* error */
	else {
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): node error\n");
		#endif	
	}

	/* update routes - only if not "self published" */
	if (src >= 0 && src <= 6 && selfpub == 0) {

		#ifdef DEBUG_BTNET_INFO
		printf("bt_node_update(): update routes hops = %d old = %d\n", new_node->routes[src].hops, new_node->routes[src].hops_old);
		#endif
	
		/* save old hop count - for later propagation */
		new_node->routes[src].hops_old = new_node->routes[src].hops;
		/* update hop count */
		if (n_type == 2) new_node->routes[src].hops = route->hops;
		if (n_type == 3) new_node->routes[src].hops = fullroute->hops;

		#ifdef DEBUG_BTNET_INFO
		printf("bt_node_update(): update routes hops = %d old = %d\n", new_node->routes[src].hops, new_node->routes[src].hops_old);
		#endif
		
		/* route deleted */
		if (new_node->routes[src].hops == 255) {
		
			/* internally no route is indicated thru "-1" */
			new_node->routes[src].hops = -1;
			
			/* change/update node state */
			if (bt_node_isconnected(new_node) == 0) {
				if (new_node->inquiry_flag == 1) {
					new_node->state = BTCHAT_NODE_STATE_VISIBLE;
				}
				else {
					new_node->state = BTCHAT_NODE_STATE_OFFLINE;
				}
				if (ret_val == 1) ret_val = 3;
			}
		
			result_node = NULL;
			res = tree_remove(&globals->btserver.clients[src].routes_learned, new_node, 0);
			#ifdef DEBUG_BTNET
			printf("bt_node_update(): routes_learned, tree_remove() = %d\n", res);
			#endif
			
			/* deleting a before unknown route, don't notify user client about existence */
			//printf("bt_node_update(): routes_learned, after tree_remove ret_val = %d\n", ret_val);
			if (ret_val == 2) ret_val = 1;
			//printf("bt_node_update(): routes_learned, after tree_remove ret_val = %d\n", ret_val);
		}
		/* route added/updated */
		else {
			result_node = NULL;
			res = tree_access_insert(&globals->btserver.clients[src].routes_learned, new_node, &result_node);
		
			switch (res) {
			case 0:	/* new node/route */
				result_node->data = new_node;
				break;
			case 1:	/* known node/route */
				/* nothing to be done! */
				break;
			default:	/* error! */
				#ifdef DEBUG_BTNET
				printf("bt_node_update(): routes_learned, res = %d, ERROR\n", res);
				#endif
				break;
			}
		}
	}
	
	/* return pointer to new/updated node */
	if (out != NULL) {
		*out = new_node;
	}
	
	/* 
	 *  if node update is because of inquiry or connect
	 *  add node to local_nodes (for later connecting)
	 */
	if (n_type == 0) {
	
		#ifdef DEBUG_BTNET
		printf("bt_node_update(): add to local_nodes\n");
		/* printf("bt_node_update(): inquiry_flag = %d\n", new_node->inquiry_flag); */
		#endif
	
		result_node = NULL;
		res = tree_access_insert(&globals->local_nodes, &work_node, &result_node);
		
		switch (res) {
		case 0:	/* new node */
			result_node->data = new_node;
			break;
		case 1:	/* known node */
			/* nothing to be done! */
			break;
		default:	/* error! */
			#ifdef DEBUG_BTNET
			printf("bt_node_update(): local_node check, res2 = %d, ERROR\n", res);
			#endif
			break;
		}

	}

	/* node is local and online but the btchat rfcomm channel is unknown - request it */
	if (new_node->state >= BTCHAT_NODE_STATE_ONLINE && ret_val == 3 &&
		new_node->discovery_method != BTCHAT_DISCOVERY_NET &&	new_node->btchat_channel == -1) {
		
		e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
		
		memset(e, 0, sizeof(struct btchat_event_t));
			
		e->etype = BTCHAT_EVENT_DEVINFO;
		e->data.node.node_ptr = new_node;
			
		/* send event to discover server */
		event_put(&globals->btdiscover.events, e);
	}

	/* 3 == state change - rewrite to general change (2) */
	if (ret_val == 3) ret_val = 2;

	/* notify user/client */
	if ((ret_val == 2 && src != 8) || selfpub == 1) {
		bt_notify_net(globals, new_node, 1);
		/* if selfpub set ret_val */
		ret_val = 2;
	}
	#ifdef DEBUG_BTNET_INFO2
	else {
		printf("bt_node_update(): bt_notify_net() ret_val = %d src = %d selfpub = %d\n", ret_val, src, selfpub);
	}
	#endif
	
	/* only propagate node information learned from other nodes */
	if (src >= 0 && src <= 6 && new_node->routes[src].hops != 0) {
		/* propagate node/route update */
		if (ret_val == 2) bt_route_propagate(globals, new_node, src, 0);
		else ret_val = bt_route_propagate(globals, new_node, src, 0);
		
		#ifdef DEBUG_BTNET_INFO2
		printf("bt_node_update(): after propagate ret_val = %d\n", ret_val);
		#endif
	}
	
	return(ret_val);
}

/**
 *  @brief notify user/net client about node status
 *
 *  @param globals pointer to globals
 *  @param node    pointer to node
 *  @param offline report offline nodes (1=yes,0=no)
 *
 *  @returns 1
 */
int bt_notify_net(struct btchatd_t *globals, struct btchat_node_t *node, int offline)
{
	struct btchat_event_t *e;
	int i;
	int hops = 255;
	int hops_index = -1;
	

	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	assert(node != NULL);
	#endif
	

	/* don't report node if offline */
	if (offline == 0 && node->state == BTCHAT_NODE_STATE_OFFLINE) {
		return(1);
	}
	
	e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
	
	#ifdef DEBUG_BTNET
	assert(e != NULL);
	#endif
	
	memset(e, 0, sizeof(struct btchat_event_t));

	/* find shortest route */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (node->routes[i].hops != -1) {
			if (hops > node->routes[i].hops) {
				hops = node->routes[i].hops;
				hops_index = i;
				if (hops == 0) break;
			}
		}
	}
	
	e->etype = BTCHAT_EVENT_NODE;
	e->prev = NULL;
	
	//memcpy(&e->data.node, node, sizeof(struct btchat_node_t));
	e->data.node.node_ptr = node;
	
	/* direct connected - online */
	if (hops == 0) {
		e->data.node.connection_type = globals->btserver.clients[hops_index].connection_direction;
		e->data.node.hops = 0;
	}
	/* node is online */
	else if (hops > 0 && hops != 255) {
		e->data.node.connection_type = BTCHAT_CONNECTION_TYPE_NET;
		e->data.node.hops = hops;
	}
	/* node isn't online */
	else {
		e->data.node.connection_type = BTCHAT_CONNECTION_TYPE_NOTCONNECTED;
		e->data.node.hops = -1;
	}
	
	/* put event to queue */
	event_put(&globals->net.events, e);
	
	return(1);
}

/**
 *  @brief send (route) the given message
 *
 *  src =   7 => network (user client)
 *        0-6 => bt socket
 *
 *  @param globals pointer to btchat globals
 *  @param src     where does the message come from
 *  @param data    pointer to message (type according to src)
 *
 *  @returns 2
 */
int bt_msg_send(struct btchatd_t *globals, int src, void *data, int data_leng)
{
	struct btcclient_protocol_msg_t *umsg;
	struct btcclient_protocol_header_t *uhe;
	struct btcnode_protocol_header_t *nhe;
	struct btcnode_protocol_msg_t *nmsg;
	struct btchat_event_t *e;
	struct btchat_node_t *node;
	struct btchat_node_t dst_node;
	struct tree_node_t *result_node;
	int dst_hops = 255;
	int dst_index = -1;
	int i;
	int res;
	#ifdef DEBUG_BTNET
	char bd_str1[20];
	#endif


	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	assert(data != NULL);
	#endif


	e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
	e->etype = BTCHAT_EVENT_MSG;

	/* from user client - outgoing to bt (allways) */
	if (src == 7) {
		#ifdef DEBUG_BTNET
		printf("bt_msg_send(): message from client to bt\n");
		#endif
	
		umsg = (struct btcclient_protocol_msg_t*) data;
		e->data.msg.msg = malloc(sizeof(struct btcnode_protocol_header_t) + data_leng);
		nhe = (struct btcnode_protocol_header_t*) e->data.msg.msg;
		nmsg = (struct btcnode_protocol_msg_t*) (e->data.msg.msg + sizeof(struct btcnode_protocol_header_t));
		
		nhe->version = 0x01;
		nhe->pkg_type = btcnode_protocol_msg;
		nhe->time = time(NULL);
	
		/* insert local bt_addr - user client doesn't know this */
		memcpy(nmsg->src, globals->self.addr.b, sizeof(nmsg->src));
		
		#ifdef DEBUG_BTNET_INFO3
		ba2str((bdaddr_t*)nmsg->src, bd_str1);
		printf("bt_msg_send(): src = %s\n", bd_str1);
		#endif
		
		memcpy(nmsg->dst, umsg->dst, sizeof(nmsg->dst));
		
		#ifdef DEBUG_BTNET_INFO3
		ba2str((bdaddr_t*)nmsg->dst, bd_str1);
		printf("bt_msg_send(): dst = %s\n", bd_str1);
		#endif
		
		memcpy((e->data.msg.msg + sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_msg_t)), (data + sizeof(struct btcclient_protocol_msg_t)), data_leng - sizeof(struct btcclient_protocol_msg_t));
		
		nhe->length = sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_msg_t) + (data_leng - sizeof(struct btcclient_protocol_msg_t));
		
		e->data.msg.msg_length = nhe->length;
		
		#ifdef DEBUG_BTNET_INFO3
		printf("bt_msg_send(): nhe->length = %d\n", nhe->length);
		#endif
		
		/* now route */
	}
	/* from bt network */
	else {
		nhe = (struct btcnode_protocol_header_t*) data;
		nmsg = (struct btcnode_protocol_msg_t*) (data + sizeof(struct btcnode_protocol_header_t));
				
		/* destination is local */
		if (memcmp(nmsg->dst, globals->self.addr.b, sizeof(nmsg->dst)) == 0) {
			e->data.msg.msg = malloc(sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_msg_t) + (nhe->length - sizeof(struct btcnode_protocol_header_t) - sizeof(struct btcnode_protocol_msg_t)));
			
			uhe = (struct btcclient_protocol_header_t*) e->data.msg.msg;
			umsg = (struct btcclient_protocol_msg_t*) (e->data.msg.msg + sizeof(struct btcclient_protocol_header_t));
			
			uhe->version = 0x01;
			uhe->pkg_type = BTCCLIENT_PROTOCOL_MSG;
			uhe->time = nhe->time;
			uhe->length = sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_msg_t) + (nhe->length - sizeof(struct btcnode_protocol_header_t) - sizeof(struct btcnode_protocol_msg_t));
			
			memcpy(umsg->src, nmsg->src, sizeof(umsg->src));
			memcpy(umsg->dst, nmsg->dst, sizeof(umsg->dst));
			memcpy((e->data.msg.msg + sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_msg_t)), (data + sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_msg_t)), (nhe->length - sizeof(struct btcnode_protocol_header_t) - sizeof(struct btcnode_protocol_msg_t)));

			#ifdef DEBUG_BTNET_INFO3
			printf("bt_msg_send(): msg length = %d => \"%s\"\n", strlen((e->data.msg.msg + sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_msg_t))), (e->data.msg.msg + sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_msg_t))); 
			#endif

			e->data.msg.msg_length = uhe->length;

			/* put to net client event queue */
			event_put(&globals->net.events, e);
			return(2);
		}
		/* destination is remote - route message from other node */
		else {
			#ifdef DEBUG_BTNET
			printf("bt_msg_send(): route message\n");
			#endif
			
			e->data.msg.msg = malloc(nhe->length);
			e->data.msg.msg_length = nhe->length;
			memcpy(e->data.msg.msg, data, nhe->length);
		}
	}

	/* search for route (device/node) */
	memset(&dst_node, 0, sizeof(struct btchat_node_t));
	memcpy(dst_node.addr.b, nmsg->dst, sizeof(dst_node.addr.b));
	result_node = NULL;
	res = tree_access(&globals->nodes, &dst_node, &result_node);
	
	node = (struct btchat_node_t*) result_node->data;
	
	/* destination found */
	if (res == 1) {
		/* find shortest route */
		for (i = 0; i < CLIENT_SLOT_NUM; i++) {
			if (node->routes[i].hops != -1) {
				if (node->routes[i].hops < dst_hops) {
					dst_hops = node->routes[i].hops;
					dst_index = i;
					if (dst_hops == 0) break;
				}
			}
		} /* for () */
		
		/* no route found */
		if (dst_index == -1) {
			#ifdef DEBUG_BTNET
			ba2str((bdaddr_t*)nmsg->dst, bd_str1);
			printf("bt_msg_send(): no route to %s!\n", bd_str1);
			#endif
			return(1);
		}
		
		#ifdef DEBUG_BTNET
		printf("bt_msg_send(): message event handed over to client[%d]\n", dst_index);
		#endif
		
		/* put message event to queue */
		event_put(&globals->btserver.clients[dst_index].events, e);
		
		return(2);
	}
	/* error */
	else {
		#ifdef DEBUG_BTNET
		ba2str((bdaddr_t*)nmsg->dst, bd_str1);
		printf("bt_msg_send(): destination = %s not found!\n", bd_str1);
		#endif
	}

	return(1);
}

/**
 *  @brief broadcast the route of given node and source to all links
 *
 *  sends routing information about given node to all links except the
 *  given link - which under normal conditions is the link over which the
 *  route was learned
 *
 *  @param globals pointer to globals
 *  @param node    pointer to node
 *  @param src     source of route
 *
 *  @returns 1
 */
int bt_route_broadcast(struct btchatd_t *globals, struct btchat_node_t *node, int src)
{
	int i;
	struct btchat_event_t e;
	struct btchat_event_t *e2;
	struct btcnode_protocol_header_t *he;
	struct btcnode_protocol_route_full_t *ro;
	#ifdef DEBUG_BTNET
	char bastr[20];
	#endif


	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	assert(node != NULL);
	#endif
	
	
	memset(&e, 0, sizeof(struct btchat_event_t));
	e.etype = BTCHAT_EVENT_ROUTE;
	he = (struct btcnode_protocol_header_t*) e.data.route.data;
	ro = (struct btcnode_protocol_route_full_t*) (e.data.route.data + sizeof(struct btcnode_protocol_header_t));

	he->version = 0x01;
	he->time = time(NULL);
	he->pkg_type = btcnode_protocol_route;
	he->length = sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_route_full_t);
	memcpy(ro->dst, node->addr.b, sizeof(ro->dst));
	if (strlen(node->alias) > 0) memcpy(ro->name, node->alias, sizeof(ro->name));
	else memcpy(ro->name, node->name, sizeof(ro->name));
	ro->state = node->state;
	if (node->routes[src].hops == -1) ro->hops = 255;
	else ro->hops = node->routes[src].hops + 1;
	
	e.data.route.data_length = he->length;
		
	#ifdef DEBUG_BTNET
	ba2str((bdaddr_t*)&ro->dst, bastr);
	printf("bt_route_broadcast(): ro_info %s [%s] state = %d\n", ro->name, bastr, ro->state);
	#endif
	
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (i != src && globals->btserver.clients[i].s != -1) {
		
			/*  TODO
			 *  implement trust_route checking!
			 *
			 */
		
			e2 = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
			memcpy(e2, &e, sizeof(struct btchat_event_t));
			
			event_put(&globals->btserver.clients[i].events, e2);
		}
	}
	
	return(1);
}

/**
 *  @brief propagate routing information about a given node
 *
 *  check the routing information of the given node, comparing the
 *  current an the previous route and if a significant change, like
 *  first route or no more routes, happens the routing information will
 *  be broadcasted to all direct connected nodes.
 *
 *  @param globals pointer to globals
 *  @param node    pointer to node
 *  @param src     source interface
 *  @param notify  notify user/net client (1=yes/0=no)
 *
 *  @returns 1 on success and 2 on success and event generated
 */
int bt_route_propagate(struct btchatd_t *globals, struct btchat_node_t *node, int src, int notify)
{
	int hops_min = 0xffff;
	int hops_index = -1;
	int i;
	

	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	assert(node != NULL);
	#endif
	
	
	/* find shortest route */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		if (node->routes[i].hops != -1) {
			if (hops_min > node->routes[i].hops) {
				hops_min = node->routes[i].hops;
				hops_index = i;
				if (hops_min == 0) break;
			}
		}
	}
	
	/* no more routes to this node */
	if (hops_index == -1) {
		/* send route delete */
		#ifdef DEBUG_BTNET
		printf("bt_route_propagate(): route deleted, propagate!\n");
		#endif
		
		/* correct node state */
		if (node->inquiry_flag == 1) node->state = BTCHAT_NODE_STATE_VISIBLE;
		else node->state = BTCHAT_NODE_STATE_OFFLINE;
		
		/* broadcast to other nodes */
		bt_route_broadcast(globals, node, src);
		
		/* notify user client */
		if (notify == 1) bt_notify_net(globals, node, 1);
		
		return(2);
	}
	/* new/updated route */
	else {
		/*  TODO
		 *  consider deletion of shortest route, routing must be updated now!
		 *  currently only new and better routes are broadcasted!
		 */
	
		/* new route is shortest */
		if (src == hops_index) {
			#ifdef DEBUG_BTNET
			printf("bt_route_propagate(): new route, propagate!\n");
			#endif
			
			/* broadcast routing information */
			bt_route_broadcast(globals, node, src);
				
			/* notify user client */
			if (notify == 1) bt_notify_net(globals, node, 1);
		
			return(2);
		}
		/* new route is NOT shortest */
		else {
			#ifdef DEBUG_BTNET
			printf("bt_route_propagate(): non-shortest route updated, ignored!\n");
			#endif
			
			return(1);
		}
	}
}

/**
 *  @brief update node and propagate changes, see bt_route_disconnected()
 *
 *  @param node pointer to node data
 *  @param data pointer to route_disconnected struct
 *
 *  @returns 1
 */
static int bt_client_disconnected_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_route_disconnected_t *d = (struct btchat_route_disconnected_t*) data;
	struct btchat_node_t *n = (struct btchat_node_t*) node->data;


	#ifdef DEBUG_BTNET
	assert(node != NULL);
	assert(data != NULL);
	#endif


	/* save old distance */
	n->routes[d->index].hops_old = n->routes[d->index].hops;
	/* set current distance to disconnected */
	n->routes[d->index].hops = -1;
	/* remember update time */
	n->routes[d->index].last_contact = d->cur_time;
	
	/* propagate route */
	bt_route_propagate(d->g, n, d->index, 1);
	
	return(1);
}

/**
 *  @brief update routing information after a client disconnected
 *
 *  update all nodes which where reachable thru the disconnecting node
 *  if a node isn't reachable no longer notify other nodes
 *
 *  @param globals pointer to globals
 *  @param index   client no. which disconnected
 *
 *  @returns 1
 */
int bt_client_disconnected(struct btchatd_t *globals, int index)
{
	struct btchat_route_disconnected_t data;


	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	#endif
	

	/* set tree walk callback function */	
	globals->btserver.clients[index].routes_learned.walk_function = bt_client_disconnected_tree_walk;
	
	/* init callback data */
	data.g = globals;
	data.index = index;
	data.cur_time = time(NULL);
	
	/* walk tree */
	tree_walk(&globals->btserver.clients[index].routes_learned, (void*)&data);
	
	/* delete tree - but not the nodes as they exist only once */
	tree_reset(&globals->btserver.clients[index].routes_learned, 0);

	return(1);
}

/**
 *  @brief tree walk callback for bt_route_publish()
 *
 *  @param node pointer to current node
 *  @param data pointer to publish data
 *
 +  @returns 1
 */
static int bt_route_publish_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_route_publish_t *d = (struct btchat_route_publish_t*) data;
	struct btchat_node_t *n = (struct btchat_node_t*) node->data;
	int i;
	int hops_min = 0xffff;
	int hops_index = -1;
	struct btchat_event_t *e;
	struct btcnode_protocol_header_t *he;
	struct btcnode_protocol_route_full_t *ro;
	#ifdef DEBUG_BTNET
	char bastr[20];
	#endif
	

	#ifdef DEBUG_BTNET
	assert(node != NULL);
	assert(data != NULL);
	#endif


	/* find shortest route */
	for (i = 0; i < CLIENT_SLOT_NUM; i++) {
		/* don't back propagate route */
		if (i != d->index) {
			/* get best */
			if (n->routes[i].hops != -1) {
				if (hops_min > n->routes[i].hops) {
					hops_min = n->routes[i].hops;
					hops_index = i;
				}
			}
		}
	}

	if (hops_index != -1) {
		e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
		memset(e, 0, sizeof(struct btchat_event_t));
		e->etype = BTCHAT_EVENT_ROUTE;
		he = (struct btcnode_protocol_header_t*) e->data.route.data;
		ro = (struct btcnode_protocol_route_full_t*) (e->data.route.data + sizeof(struct btcnode_protocol_header_t));

		#ifdef DEBUG_BTNET
		ba2str(&n->addr, bastr);
		printf("bt_route_publish_tree_walk(): addr = %s name = %s hops = %d\n", bastr, n->name, n->routes[hops_index].hops + 1);
		#endif

		he->version = 0x01;
		he->time = d->cur_time;
		he->pkg_type = btcnode_protocol_route_full;
		he->length = sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_route_full_t);
		memcpy(ro->dst, n->addr.b, sizeof(ro->dst));
		if (strlen(n->alias) > 0) memcpy(ro->name, n->alias, sizeof(ro->name));
		else memcpy(ro->name, n->name, sizeof(ro->name));
		ro->state = n->state;
		ro->hops = n->routes[hops_index].hops + 1;
	
		e->data.route.data_length = he->length;
				
		#ifdef DEBUG_BTNET
		ba2str((bdaddr_t*)&ro->dst, bastr);
		printf("bt_route_publish_tree_walk(): ro_info %s [%s] hops = %d\n", ro->name, bastr, ro->hops);
		#endif

		event_put(&d->g->btserver.clients[d->index].events, e);
	}
	
	return(1);
}

/**
 *  @brief publish all known routes
 *
 *  this sends all known routes to a connected client/node, this normally happens
 *  on a new connection after trust levels haven been raised to allow routing
 *
 *  @param globals pointer to globals
 *  @param index   index of client
 *
 *  @returns 1
 */
int bt_route_publish(struct btchatd_t *globals, int index)
{
	void *old_walk_function;
	struct btchat_route_publish_t data;


	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	#endif
	
	
	old_walk_function = globals->nodes.walk_function;
	globals->nodes.walk_function = bt_route_publish_tree_walk;
	
	data.g = globals;
	data.index = index;
	data.cur_time = time(NULL);
	
	tree_walk(&globals->nodes, (void*)&data);
	
	globals->nodes.walk_function = old_walk_function;
	
	return(1);
}

/**
 *  @brief publish information about ourself
 *
 *  @param globals pointer to globals
 *  @param index   index of connection
 *
 *  @returns 2
 */
int bt_self_publish(struct btchatd_t *globals, int index)
{
	struct btchat_event_t *e;
	struct btcnode_protocol_header_t *he;
	struct btcnode_protocol_route_full_t *ro;
	#ifdef DEBUG_BTNET
	char bastr[20];
	#endif
	

	#ifdef DEBUG_BTNET
	assert(globals != NULL);
	#endif


	e = (struct btchat_event_t*) malloc(sizeof(struct btchat_event_t));
	memset(e, 0, sizeof(struct btchat_event_t));
	e->etype = BTCHAT_EVENT_ROUTE;
	he = (struct btcnode_protocol_header_t*) e->data.route.data;
	ro = (struct btcnode_protocol_route_full_t*) (e->data.route.data + sizeof(struct btcnode_protocol_header_t));

	he->version = 0x01;
	he->time = time(NULL);
	he->pkg_type = btcnode_protocol_route_full;
	he->length = sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_route_full_t);
	memcpy(ro->dst, globals->self.addr.b, sizeof(ro->dst));
	if (strlen(globals->self.alias) > 0) memcpy(ro->name, globals->self.alias, sizeof(ro->name));
	else memcpy(ro->name, globals->self.name, sizeof(ro->name));
	ro->state = 0; /* ignored anyway */
	ro->hops = 0; /* correct but ignored */
	
	e->data.route.data_length = he->length;
	
	#ifdef DEBUG_BTNET
	ba2str((bdaddr_t*)&ro->dst, bastr);
	printf("bt_self_publish(): %s [%s] hops = %d\n", ro->name, bastr, ro->hops);
	#endif

	event_put(&globals->btserver.clients[index].events, e);

	return(2);
}

/*
 *  $Log$
 */
