/*
 *  $Id$
 */

/**
 *  @file btchatd.h
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#include "node_protocol.h"

/**
 *  @brief number of client slots
 * 
 *  bluetooth supports up to 7 active connections
 */
#define CLIENT_SLOT_NUM 7

#ifndef __BTCHATD_H__
#define __BTCHATD_H__

#ifndef __BTCHAT_ENUMS__
#define __BTCHAT_ENUMS__

enum trust_level_route_t {
	BTCHAT_TRUST_LEVEL_ROUTE_DONTCONNECT = -1, /**< don't allow connections */
	BTCHAT_TRUST_LEVEL_ROUTE_NEUTRAL = 0,      /**< don't allow routing */
	BTCHAT_TRUST_LEVEL_ROUTE_ALLOW             /**< route traffic */
};

enum trust_level_general_t {
	BTCHAT_TRUST_LEVEL_GENERAL_IGNORE = -1, /**< ignore user/node */
	BTCHAT_TRUST_LEVEL_GENERAL_NEUTRAL = 0, /**< normal */
	BTCHAT_TRUST_LEVEL_GENERAL_FRIEND       /**< user/node is a friend */
};

enum discovery_method_t {
	BTCHAT_DISCOVERY_NULL = 0,
	BTCHAT_DISCOVERY_INQUIRY,   /**< node is found thru an inquiry */
	BTCHAT_DISCOVERY_CONNECT,   /**< node is found thru in incoming connection */
	BTCHAT_DISCOVERY_NET,       /**< node is found thru the net */
	BTCHAT_DISCOVERY_LAST = 255
};

enum connection_direction_t {
	BTCHAT_CONNECTION_NOTCONNECTED = 0, /**< no connection */
	BTCHAT_CONNECTION_INCOMING = 1,     /**< incoming connection */
	BTCHAT_CONNECTION_OUTGOING = 2      /**< outgoing connection */
};

enum btchat_node_direct_state_t {
	BTCHAT_NODE_DIRECT_STATE_DISCONNECTED = 0,
	BTCHAT_NODE_DIRECT_STATE_CONNECTED,
	BTCHAT_NODE_DIRECT_STATE_DISCONNECT,
	BTCHAT_NODE_DIRECT_STATE_CONNECT
};

enum btchat_node_state_t {
	BTCHAT_NODE_STATE_FIRST = 0,
	BTCHAT_NODE_STATE_OFFLINE,   /**< not visible/connected */
	BTCHAT_NODE_STATE_VISIBLE,   /**< seen thru a bluetooth inquiry */
	BTCHAT_NODE_STATE_ONLINE,    /**< somehow connected to the net */
	BTCHAT_NODE_STATE_AWAY,      /**< connected but idle/away */
	BTCHAT_NODE_STATE_LAST = 255
};

enum btchat_route_policy_t {
	BTCHAT_ROUTE_POLICY_FIRST = 0,
	BTCHAT_ROUTE_POLICY_NEVER,     /**< do not route any traffic */
	BTCHAT_ROUTE_POLICY_ALLWAYS,   /**< allways route */
	BTCHAT_ROUTE_POLICY_TRUSTED,	 /**< only route traffic from trusted devices - default */
	BTCHAT_ROUTE_POLICY_NEEDUSER,	 /**< only route if user is active */
	BTCHAT_ROUTE_POLICY_LAST = 255
};
#endif

struct btchat_route_dst_t {
	int hops;            /**< number of hops to destination, 0 = direct connection, -1 = not route */
	int hops_old;        /**< old hops of this link */
	time_t last_contact; /**< route updated */
};

struct btchat_node_t {

	bdaddr_t addr;             /**< bluetooth device address */
	unsigned int device_class; /**< bluetooth device service class */
	char name[250];            /**< bluetooth device name */	
	short int btchat_channel;  /**< btchat rfcomm channel (-1 = not available) */

	unsigned char discovery_method; /**< method of device discovery */
	unsigned char inquiry_flag;     /**< device is locally visible thru BT inquiry */
	int state;                      /**< status of the node */
		
	char alias[250]; /**< the visible name */
	
	/* local to remote (set by local user) */
	int trust_route;   /**< routing trust level */
	int trust_general; /**< general trust level */
	
	/* remote to local (set by remote user) */
	int trusted_route;   /**< routing trust level */
	int trusted_general; /**< general trust level */
	
	struct btchat_route_dst_t routes[CLIENT_SLOT_NUM]; /**< routing information */
	int routes_shortest;                               /**< index of shortest route */
	time_t contact_last;                               /**< time stamp of last contact to node */
};

enum btchat_event_type_t {
	BTCHAT_EVENT_NULL = 0,
	BTCHAT_EVENT_DISCOVER,
	BTCHAT_EVENT_DEVINFO,
	BTCHAT_EVENT_NODE,
	BTCHAT_EVENT_MSG,
	BTCHAT_EVENT_ROUTE,
	BTCHAT_EVENT_LASTEVENT = 255
};

/**
 *  @brief a btchat event
 *
 */
struct btchat_event_t {

	int etype; /**< event type */
	
	union {
		struct {
			struct btchat_node_t node;
			struct btchat_node_t *node_ptr;
			int connection_type;
			int hops;
		} node;
		struct {
			unsigned char *msg;
			int msg_length;
		} msg;
		struct {
			int data_length;
			unsigned char data[sizeof(struct btcnode_protocol_header_t) + sizeof(struct btcnode_protocol_route_full_t)];
		} route;
	} data;
	
	struct btchat_event_t *prev; /**< previous event in queue */
};

/**
 *  @brief event queue
 *
 */
struct btchat_event_queue_t {

	struct btchat_event_t *head; /**< first (newest) event in queue */
	struct btchat_event_t *tail; /**< last (oldest) event in queue */
};

struct btchat_btdiscover_client_t {

	int s; /**< socket */
	
	int pid; /**< the child pid */
	
	int inquiry_duration;       /**< bt inquiry duration */
	int local_monitor;          /**< yes=1 no=0 */
	int local_monitor_interval; /**< inquiry every X seconds */
	time_t local_monitor_last;  /**< last inquiry time */
	
	unsigned char buffer[0xffff]; /**< receive buffer */
	unsigned int buffer_pos;      /**< position in buffer */
	unsigned int buffer_size;     /**< size of buffer */
	
	struct btchat_event_queue_t events; /**< event queue */
};

struct btchat_btdirect_t {

	struct btchat_node_t *node; /**< the actual node */
	
	int connection_direction;   /**< role in connection, incoming or outgoing */
	
	int s;                 /**< the connection socket */
	struct sockaddr_rc sa; /**< the socket info */
	int sa_len;            /**< socket info struct size */
	
	unsigned char buffer[0xffff]; /**< read buffer */
	unsigned int buffer_pos;      /**< position in read buffer */
	unsigned int buffer_size;     /**< size of read buffer */
	
	int state;          /**< connection state */	
	int selector_index; /**< index for global event_loop */

	struct tree_t routes_learned; /**< routes learned over this link */
	
	struct btchat_event_queue_t events; /**< event queue */
	
	unsigned long int bytes_send; /**< statistics bytes send */
	unsigned long int bytes_recv; /**< statistics bytes received */
};

struct btchat_btserver_t {

	sdp_record_t *sdp_record;   /**< btChat SDP record */
	sdp_session_t *sdp_session; /**< local SDP session */

	int server;                   /**< server socket */
	struct sockaddr_rc server_sa; /**< socket info */
	int server_len;               /**< socket info size */

	struct btchat_btdirect_t clients[CLIENT_SLOT_NUM]; /**< direct connected clients */
	int clients_num;                                   /**< current number of clients */
	int clients_max;                                   /**< max number of clients to connect */

	time_t connect_last;      /**< time at which the last outgoing connection was made */
	int connect_delay;        /**< delay between connection attempts */
	int connect_min_distance; /**< minimal distance all ready connected nodes */
	
	int state; /**< current server state */
	
	struct btchat_event_queue_t events; /**< event queue */
	
	int routing_policy; /**< current routing policy */
};

struct btchat_network_t {

	short int port;     /**< TCP port to listen on */
	char local_ip[256]; /**< ip to bind to */

	struct hostent *hostinfo;     /**< hostinfo for bind() */
	struct sockaddr_in laddr;     /**< local address */
	struct sockaddr_in remote_sa; /**< client socket information */
	socklen_t remote_len;         /**< size of remote_sa */

	int server; /**< server socket */
	int client; /**< client socket */

	unsigned char buffer[0xffff]; /**< read buffer */
	unsigned int buffer_pos;      /**< read buffer position */
	unsigned int buffer_size;     /**< read buffer size */

	int state; /**< network server state */
	
	struct btchat_event_queue_t events; /**< event queue */
};

struct btchat_cron_t {

	int interval; /**< delay between tasks */
	
	time_t next;  /**< time() of next task */
	time_t last;  /**< time() of last task */
};

struct btchatd_t {
	struct btchat_node_t self; /**< local node */
	
	struct btchat_network_t net;                  /**< local network, user client */
	struct btchat_btserver_t btserver;            /**< bluetooth remote nodes/users */
	struct btchat_btdiscover_client_t btdiscover; /**< bluetooth device discovery */

	struct tree_t nodes;       /**< all known nodes */
	struct tree_t local_nodes; /**< neighbor nodes, the data pointer points to
	                            *   the data in: nodes 
	                            */

	struct tree_t bcast; /**< current broadcasts */
	int bcast_timeout;   /**< broadcast timeout */
	
	struct btchat_cron_t tasks;

	int running; /**< global running flag */
};

#endif

/*
 *  $Log$
 */
