/*
 *  $Id$
 */
 
/**
 *  @file event.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "event.h"
#include "bt_discover.h"
#include "bt_net.h"

/**
 *  @brief btchatd main event loop
 *
 *  @param globals pointer to globals
 *
 *  @returns 1
 */
int event_loop(struct btchatd_t *globals)
{
	struct pollfd selectors[11];
	int selectors_num;
	int selectors_ready;
	int btserver_result;
	int netserver_result;
	int btdiscover_result;
	int btclient_result;
	int i;
	int event_delay = 0;
	time_t tmp_time;
	struct btchat_event_t *e;
	
	
	#ifdef DEBUG_EVENT
	assert(globals != NULL);
	#endif
	
	
	while (globals->running == 1) {
		memset(selectors, 0, sizeof(selectors));
		selectors_num = 0;
		selectors_ready = 0;
		btserver_result = 0;
		netserver_result = 0;
		btdiscover_result = 0;
		btclient_result = 0;
		

		/* user client */		
		if (globals->net.client == -1) selectors[0].fd = globals->net.server;
		else selectors[0].fd = globals->net.client;
		selectors[0].events = POLLIN | POLLPRI;
		selectors_num++;

		/* bluetooth server */
		selectors[1].fd = globals->btserver.server;
		selectors[1].events = POLLIN | POLLPRI;
		selectors_num++;

		/* device discovery */
		selectors[2].fd = globals->btdiscover.s;
		selectors[2].events = POLLIN | POLLPRI;
		selectors_num++;

		/* connected bluetooth clients/nodes */
		for (i = 0; i < CLIENT_SLOT_NUM; i++) {
			if (globals->btserver.clients[i].s != -1) {
				selectors[selectors_num].fd = globals->btserver.clients[i].s;
				selectors[selectors_num].events = POLLIN | POLLPRI;
				globals->btserver.clients[i].selector_index = selectors_num;
				selectors_num++;
			}
		}

		/* poll all descriptors */
		#ifdef DEBUG_EVENT
		printf("event_loop(): event_delay = %dsec\n", event_delay/1000);
		#endif
		selectors_ready = poll(selectors, selectors_num, event_delay);
		#ifdef DEBUG_EVENT
		perror("event_loop()");
		#endif
		
		/* sleep forever */
		event_delay = -1;
		
		/* user client (network) */	
		netserver_result = net_server(globals, &selectors[0]);

		/* incoming bluetooth connections */
		btserver_result = bt_server(globals, &selectors[1]);

		/* bluetooth device discovery */
		btdiscover_result = bt_discover_client(globals, &selectors[2]);

		/* active bluetooth connections */
		for (i = 0; i < CLIENT_SLOT_NUM; i++) {
			if (globals->btserver.clients[i].selector_index != -1) {
				btclient_result = bt_client(globals, &selectors[globals->btserver.clients[i].selector_index], i);
				if (btclient_result > 1) {
					#ifdef DEBUG_EVENT
					printf("event_loop(): bt_client\n");
					#endif
					event_delay = 0;
				}
			}
		}
		
		/* do we need to make out going bluetooth connections? */
		btclient_result = bt_client_connect(globals);

		/* can we go to sleep or do we need to finish pending events */
		if (netserver_result > 1) {
			#ifdef DEBUG_EVENT
			printf("event_loop(): net\n");
			#endif
			event_delay = 0;
		}
		else if (btserver_result > 1) {
			#ifdef DEBUG_EVENT
			printf("event_loop(): bt_server\n");
			#endif
			event_delay = 0;
		}
		else if (btdiscover_result > 1) {
			#ifdef DEBUG_EVENT
			printf("event_loop(): bt_discovery\n");
			#endif
			event_delay = 0;
		}
		
		/* cron jobs */
		if (event_delay != 0) {

			if (globals->tasks.interval > 0) {
				time(&tmp_time);
				
				if (tmp_time >= globals->tasks.next) {
					/* do task */
					
					/* start device discovery */
					e = malloc(sizeof(struct btchat_event_t));
					memset(e, 0, sizeof(struct btchat_event_t));
					e->etype = BTCHAT_EVENT_DISCOVER;
					event_put(&globals->btdiscover.events, e);
			
					time(&globals->tasks.next);
					
					/*  increase the delay between inquiries with each connected
					 *  node to minimize interference
					 *
					 *  interval * (number of connections + 1) early increase delay between inquires
					 */
					globals->tasks.next += (globals->tasks.interval * (globals->btserver.clients_num + 1));
					
					event_delay = 0;
				}
				else {
					/* calc. sleep time */
					event_delay = (globals->tasks.next - tmp_time) * 1000;
				}
			}
		}

	} /* while (globals->running == 1) */
	
	return(1);
}

/**
 *  @brief insert an event into a given queue
 *
 *  @param q pointer to event queue
 *  @param e pointer to event
 *
 *  @returns 1 on success and -1 on error
 */
int event_put(struct btchat_event_queue_t *q, struct btchat_event_t *e)
{
	int result = 1;
	

	#ifdef DEBUG_EVENT
	assert(q != NULL);
	assert(e != NULL);
	#endif


	if (q->head == NULL) {
		q->head = e;
		q->tail = e;
		e->prev = NULL;
	}
	else if (q->head != NULL) {
		e->prev = NULL;
		q->head->prev = e;
		q->head = e;
	}
	else {
		#ifdef DEBUG_EVENT
		printf("event_put(): error\n");
		#endif
		q->head = NULL;
		q->tail = NULL;
		result = -1;
	}
	
	return(result);
}

/**
 *  @brief get next event from event queue
 *
 *  @param q pointer to event queue
 *  @param e pointer to pointer where to store the event
 *
 *  @returns 1 on success and 0 on error
 */
int event_get(struct btchat_event_queue_t *q, struct btchat_event_t **e)
{
	int retval = 1;


	#ifdef DEBUG_EVENT
	assert(q != NULL);
	assert(e != NULL);
	#endif


	if (q->tail != NULL) {
		*e = q->tail;
		q->tail = (*e)->prev;
		if (q->tail == NULL) q->head = NULL;
	}
	else {
		retval = 0;
	}
	
	return(retval);
}

/**
 *  @brief check if event queue contains events
 *
 *  @param q pointer to event queue
 *
 *  @returns 1 if events are available and 0 if not
 */
int event_available(struct btchat_event_queue_t *q)
{
	#ifdef DEBUG_EVENT
	assert(q != NULL);
	#endif

	if (q->tail != NULL) return(1);
	else return(0);
}

/**
 *  @brief reset/clear event queue
 *
 *  @param q pointer to event queue
 *
 *  @returns 1
 */
int event_queue_reset(struct btchat_event_queue_t *q)
{
	struct btchat_event_t *tmp;
	

	#ifdef DEBUG_EVENT
	assert(q != NULL);
	#endif
	
	
	while (event_get(q, &tmp) == 1) {
		event_free(&tmp);
	}
	
	return(1);
}

/**
 *  @brief free memory used by an event
 *
 *  @param e pointer to pointer to event
 *
 *  @returns 1
 */
int event_free(struct btchat_event_t **e)
{
	#ifdef DEBUG_EVENT
	assert(e != NULL);
	#endif
	
	
	switch ((*e)->etype) {
	case BTCHAT_EVENT_MSG:
		if ((*e)->data.msg.msg != NULL) free((*e)->data.msg.msg);
		break;
	};
	
	free(*e);
	*e = NULL;

	return(1);
}

/*
 *  $Log$
 */
