/*
 *  $Id$
 */

/**
 *  @file main.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "event.h"
#include "net.h"
#include "bt.h"
#include "util.h"
#include "sdp.h"
#include "bt_discover.h"

#define BTCHAT_INFO_TEXT ""\
"\nWelcome to btChat - bluetooth adhoc communication!\n\n"\
"GAIM setup:\n"\
"\tScreen Name:\tyour nick\n"\
"\tPassword:\tanything\n"\
"\tProxy type:\tNo Proxy\n"\
"\n"\
"btChatd version: 0.1.1\n"\
"contact: collin-btchat@betaversion.net\n"\
"web:     http://www.mulliner.org/bluetooth/btchat/\n\n"\
"start GAIM and connect now\n\n"

/* global data */
struct btchatd_t *globals = NULL;
int hci_dev;

void usage(char *prgname)
{
	printf(	"%s -i <hciX> -r <channel> -c <max connections>\n"\
				"\tchannel accepts 1-30 and defaults to auto select\n"\
				"\tmax connections accepts 1-7 and defaults to 5\n", prgname);
}

/**
 *  @brief signal handler for SIG_TERM
 *
 *  @param signum signal number
 *
 *  @returns nothing
 */
void signal_term(int signum)
{
	if (globals->net.server != -1) close(globals->net.server);
	if (globals->net.client != -1) close(globals->net.client);
	
	/* stop event-loop */
	globals->running = 0;
}

/**
 *  @brief btchatd main()
 *
 *  @param argc number of arguments
 *  @param argv argument array
 *
 *  @returns 0
 */
int main(int argc, char **argv)
{
	int commsockets[2];
	int childpid;
	int i;
	int opt;
	int hci_dd;
	unsigned int max_con = 5;
	unsigned int rf_chan = 0;
	int expert = 0;
	char hci_if[5] = {"hci0\0"};
	#ifdef DEBUG_MAIN
	char ba_str[20];
	#endif


	while ((opt = getopt(argc, argv, "i:c:r:e")) != EOF) {
		switch (opt) {
		case 'i':
			strncpy(hci_if, optarg, 5);
			break;
		case 'c':
			max_con = atoi(optarg);
			break;
		case 'r':
			rf_chan = atoi(optarg);
			break;			
		case 'e':
			expert = 1;
			break;
		default:
			usage(argv[0]);
			exit(-1);
		}
	}
	
	if ((hci_dev = hci_devid(hci_if)) < 0) {
		printf("%s: interface invalid\n", hci_if);
		usage(argv[0]);
		exit(-1);
	}
	if (max_con < 1 || max_con > 7) {
		printf("%d: wrong number of maximal connections\n", max_con);
		usage(argv[0]);
		exit(-1);
	}
	if (rf_chan > 30) {
		printf("%d: invalid rfcomm channel\n", rf_chan);
		usage(argv[0]);
		exit(-1);
	}
	
	if (expert == 0) {
		printf(BTCHAT_INFO_TEXT);
		sleep(1);
	}

	//signal(SIGINT, signal_term);
	signal(SIGTERM, signal_term);
	
	/* bt_discover server */
	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, commsockets) != 0) {
		#ifdef DEBUG_MAIN
		perror("main(): socketpair");
		#endif
		return(-1);
	}
	
	/* fork discover process */
	childpid = fork();
	if (childpid == -1) {
		#ifdef DEBUG_MAIN
		perror("main(): fork");
		#endif
		return(-1);
	}
	/* this is the child */
	else if (childpid == 0) {
		close(commsockets[1]);
		bt_discover_server(commsockets[0]);
		return(1);
	}
	/* this is the main process */
	else {
		close(commsockets[0]);
		
		globals = (struct btchatd_t*)	malloc(sizeof(struct btchatd_t));
		#ifdef DEBUG_MAIN
		assert(globals != NULL);
		#endif
		memset(globals, 0, sizeof(struct btchatd_t));
		
		/* bt_discover_client() setup */
		globals->btdiscover.s = commsockets[1];
		globals->btdiscover.pid = childpid;
		globals->btdiscover.inquiry_duration = 10;
		globals->btdiscover.buffer_pos = 0;
		globals->btdiscover.buffer_size = 0xffff;
		globals->btdiscover.local_monitor = 1;
		globals->btdiscover.local_monitor_interval = 60;
		globals->btdiscover.local_monitor_last = 0;
	
		/* net_server() setup */
		globals->net.port = 2222;
		strncpy(globals->net.local_ip, "127.0.0.1\0", 256);
		globals->net.server = -1;
		globals->net.client = -1;
		globals->net.state = 0;
		globals->net.buffer_pos = 0;
		globals->net.buffer_size = 0xffff;
	
		/* bt_server() setup */
		globals->btserver.server = -1;
		globals->btserver.state = 0;
		globals->btserver.clients_num = 0;
		globals->btserver.clients_max = max_con;
		globals->btserver.routing_policy = BTCHAT_ROUTE_POLICY_ALLWAYS;
		//globals->btserver.routing_policy = BTCHAT_ROUTE_POLICY_TRUSTED;
		globals->btserver.connect_delay = 10;
		globals->btserver.connect_min_distance = 3; /* not lower then 2 */
		
		globals->self.btchat_channel = rf_chan;
		/* read local device name */
		hci_dd = hci_open_dev(hci_dev);
		hci_read_local_name(hci_dd, sizeof(globals->self.name), globals->self.name, 100000);
		hci_close_dev(hci_dd);
		/* get local device bd_addr */
		hci_devba(hci_dev, &globals->self.addr);
		#ifdef DEBUG_MAIN
		ba2str(&globals->self.addr, ba_str);
		printf("main(): local = %s [%s]\n", globals->self.name, ba_str);
		#endif
		
		/* init direct connections */
		for (i = 0; i < CLIENT_SLOT_NUM; i++) {
			globals->btserver.clients[i].s = -1;
			globals->btserver.clients[i].selector_index = -1;
			globals->btserver.clients[i].buffer_pos = 0;
			globals->btserver.clients[i].buffer_size = 0xffff;
			globals->btserver.clients[i].connection_direction = 0;
			
			globals->btserver.clients[i].routes_learned.splay = 1;
			globals->btserver.clients[i].routes_learned.num = 0;
			globals->btserver.clients[i].routes_learned.root = NULL;
			globals->btserver.clients[i].routes_learned.compare_function = btchat_tree_cmp;
			globals->btserver.clients[i].routes_learned.walk_function = NULL;
		}
		
		if (net_server(globals, NULL) != 1) {
			#ifdef DEBUG_MAIN
			printf("main(): net_server failed\n");
			#endif
			return(-1);
		}
		
		if (bt_server(globals, NULL) != 1) {
			#ifdef DEBUG_MAIN
			printf("main(): bt_server failed\n");
			#endif
			return(-1);
		}

		/* node tree */
		globals->nodes.splay = 1;
		globals->nodes.num = 0;
		globals->nodes.root = NULL;
		globals->nodes.compare_function = btchat_tree_cmp;
		globals->nodes.walk_function = NULL;
		
		/* local_node tree */
		globals->local_nodes.splay = 1;
		globals->local_nodes.num = 0;
		globals->local_nodes.root = NULL;
		globals->local_nodes.compare_function = btchat_tree_cmp;
		globals->local_nodes.walk_function = NULL;
		
		/* general setup */
		globals->running = 1;
		
		/* "cron" jobs */
		globals->tasks.interval = 10; /* seconds */
		globals->tasks.last = 0;
		globals->tasks.next = 0;
						
		/* application main loop */
		return(event_loop(globals));
	}
	
	return(0);
}

/*
 *  $Log$
 */
