/*
 *  $Id$
 */

/**
 *  @file net.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "event.h"
#include "util.h"
#include "client_protocol.h"
#include "net.h"
#include "bt.h"
#include "bt_net.h"

/**
 *  @brief handle the connection with the user front-end
 *
 *  This function binds to the given local ip-address and port, normally
 *  127.0.0.1 and accepts exactly one (1) connection.
 *
 *  @param args pointer to globals
 *  @param fd   pointer to file descriptor
 *
 *  @returns 1 on success and -1 on failure
 */
int net_server(struct btchatd_t *globals, struct pollfd *fd)
{
	int read_bytes;
	int write_bytes;
	int ret_val = 1;
	struct btchat_event_t *event;
	unsigned char *out_buffer = NULL;
	unsigned int out_buffer_size;
	unsigned int out_buffer_pos;
	struct btcclient_protocol_account_t *dev;
	struct btcclient_protocol_header_t *he;
	struct btcclient_protocol_msg_t *msg;
	struct btcclient_protocol_login_t *login;
	#ifdef DEBUG_NET_INFO
	char bdstr[20];
	#endif
	

	#ifdef DEBUG_NET
	assert(globals != NULL);
	#endif
	
	
	/*
	 *	state == 0 => network initialization
	 */
	if (globals->net.state == 0) {
	
		globals->net.laddr.sin_port = htons((u_short) globals->net.port);
		if ((globals->net.hostinfo = gethostbyname(globals->net.local_ip))) {
			bcopy(globals->net.hostinfo->h_addr, (char *) &globals->net.laddr.sin_addr, globals->net.hostinfo->h_length);
			globals->net.laddr.sin_family = AF_INET;
		}
		else {
			#ifdef DEBUG_NET
			perror("net_server(): gethostbyname failed");
			#endif
			return(-1);
		}
	
		if ((globals->net.server = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
			#ifdef DEBUG_NET
			perror("net_server(): socket failed");
			#endif
			return(-1);
		}
	
		if (bind(globals->net.server, (struct sockaddr*) &globals->net.laddr, sizeof(struct sockaddr_in)) == -1) {
			#ifdef DEBUG_NET
			perror("net_server(): bind failed");
			#endif
			return(-1);
		}
	
		if (listen(globals->net.server, 1) == -1) {
			#ifdef DEBUG_NET
			perror("net_server(): listen failed");
			#endif
			return(-1);
		}
		
		globals->net.state = 1;
		return(1);
	} /* state == 0 */
	/*
	 *	state == 1 => accepting new client connection
	 */
	else if (globals->net.state == 1) {
		if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
		
			/* accept incoming client connection */
			if (fd->fd == globals->net.server) {
				globals->net.client = accept(globals->net.server, (struct sockaddr*) &globals->net.remote_sa, &globals->net.remote_len);
				
				/* close server socket */
				close(globals->net.server);
				globals->net.server = -1;
				
				/* we have a client now */
				globals->net.state = 2;
				
				globals->net.buffer_pos = 0;
				
				#ifdef DEBUG_NET_INFO
				printf("net_server(): accepted new connection\n");
				#endif
				
				net_notify_con(globals);
				
				#ifdef DEBUG_NET_INFO
				printf("net_server(): sent current nodes to client\n");
				#endif
			}
			else {
				#ifdef DEBUG_NET
				printf("net_server(): server socket corrupted, going to state = 0\n");
				#endif
								
				goto disconnected;
			}
		}
	}
	/*
	 *	state == 2 => talk to client
	 */
	else if (globals->net.state == 2) {
		if (fd->fd == globals->net.client) {
			if ((fd->revents & POLLIN) == POLLIN || (fd->revents & POLLPRI) == POLLPRI) {
			
				read_bytes = cread(globals->net.client, globals->net.buffer + globals->net.buffer_pos, globals->net.buffer_size - globals->net.buffer_pos);
				
				/* client disconnected */
				if (read_bytes == 0) {
					#ifdef DEBUG_NET_INFO
					printf("net_server(): user client has closed the connection\n");
					#endif
					
					goto disconnected;
				}
				
				globals->net.buffer_pos += read_bytes;
				
				/* do protocol action here ! */
			check_buffer_again:
				if (globals->net.buffer_pos >= sizeof(struct btcclient_protocol_header_t )) {
					he = (struct btcclient_protocol_header_t*) globals->net.buffer;
				
					#ifdef DEBUG_NET_INFO
					printf("net_server(): he->length = %d, buffer_pos = %d\n", he->length, globals->net.buffer_pos);
					#endif
				
					if (globals->net.buffer_pos >= he->length) {

						/* message packet */
						if (he->pkg_type == BTCCLIENT_PROTOCOL_MSG) {
							msg = (struct btcclient_protocol_msg_t*) (globals->net.buffer + sizeof(struct btcclient_protocol_header_t));
							
							ret_val = bt_msg_send(globals, 7, (globals->net.buffer + sizeof(struct btcclient_protocol_header_t)), (he->length - sizeof(struct btcclient_protocol_header_t)));
						}
						/* login packet */
						else if (he->pkg_type == BTCCLIENT_PROTOCOL_LOGIN) {
							login = (struct btcclient_protocol_login_t*) (globals->net.buffer + sizeof(struct btcclient_protocol_header_t));
							
							memcpy(globals->self.alias, login->name, sizeof(globals->self.alias));
							
							#ifdef DEBUG_NET_INFO
							printf("net_server(): login %s\n", login->name);
							#endif
						}
						/* unknown packet type */
						else {
							/*  TODO
							 *  tidy up buffer here
							 */
							#ifdef DEBUG_NET
							printf("net_server(): unknown packet type!\n");
							#endif
						}
						
						/* tidy up read buffer */
						memmove(globals->net.buffer, (globals->net.buffer + he->length), (globals->net.buffer_pos - he->length));
						globals->net.buffer_pos -= he->length;
						
						/* check buffer again */
						goto check_buffer_again;

					}
					/* packet not complete */
					else {
					}
				}
				/* header not complete */
				else {
				}
			
			} /* POLLIN || POLLPRI */
			else {
			}
		}
		else {
			#ifdef DEBUG_NET
			printf("net_server(): fd->fd has strange value\n");
			printf("net_server(): closing connection\n");
			#endif
			
			goto disconnected;
		}

		/* process pending events */		
		if (event_get(&globals->net.events, &event) == 1) {
		
			/* node information event */
			if (event->etype == BTCHAT_EVENT_NODE) {
				out_buffer_size = sizeof(struct btcclient_protocol_header_t) + sizeof(struct btcclient_protocol_account_t);
				out_buffer_pos = 0;
				out_buffer = (unsigned char*) malloc(out_buffer_size);
				
				#ifdef DEBUG_NET
				assert(out_buffer != NULL);
				#endif
				
				memset(out_buffer, 0, out_buffer_size);
				
				he = (struct btcclient_protocol_header_t*) out_buffer;
				dev = (struct btcclient_protocol_account_t*) (out_buffer + sizeof(struct btcclient_protocol_header_t));
				
				he->version = 0x01;
				he->pkg_type = BTCCLIENT_PROTOCOL_ACCOUNT;
				time((time_t*) &he->time);
				he->length = out_buffer_size;
				
				memcpy(dev->addr, event->data.node.node_ptr->addr.b, 6);
				dev->device_class = event->data.node.node_ptr->device_class;
				
				if (strlen(event->data.node.node_ptr->alias) > 0) memcpy(dev->name, event->data.node.node_ptr->alias, sizeof(dev->name));
				else memcpy(dev->name, event->data.node.node_ptr->name, sizeof(dev->name));
				
				dev->discovery_method = event->data.node.node_ptr->discovery_method;

				dev->trust_route = event->data.node.node_ptr->trust_route;
				dev->trust_general = event->data.node.node_ptr->trust_general;
				
				dev->connection = event->data.node.connection_type;
				if (event->data.node.hops == -1) dev->hops = 255; 
				else dev->hops = event->data.node.hops;
				
				dev->state = event->data.node.node_ptr->state;
				
				#ifdef DEBUG_NET_INFO
				ba2str(&event->data.node.node.addr, bdstr);
				printf("net_server(): node %s [%s] state = %d con = %d hops = %d\n", dev->name, bdstr, dev->state, dev->connection, dev->hops);
				#endif
			}
			else if (event->etype == BTCHAT_EVENT_MSG) {
				write_bytes = nwrite(globals->net.client, event->data.msg.msg, event->data.msg.msg_length);
			}
			/* unknown event - free and return */
			else {
				event_free(&event);
				
				/* check pending events */
				goto pending_events;
			}
			
			if (out_buffer != NULL) {
				write_bytes = nwrite(globals->net.client, out_buffer, out_buffer_size);
			}
			#ifdef DEBUG_NET_INFO
			printf("net_server(): write_bytes = %d\n", write_bytes);
			#endif
			
			if (write_bytes == 0) {
				event_free(&event);
				if (out_buffer != NULL) free(out_buffer);
				out_buffer = NULL;
				
				#ifdef DEBUG_NET_INFO
				printf("net_server(): user client has closed the connection\n");
				#endif
	
				goto disconnected;
			}
		
			/* cleanup */
			if (out_buffer != NULL) free(out_buffer);
			out_buffer = NULL;
			event_free(&event);
		}
	} /* state == 2 */
	
	/* clear event queue - no client to consume events */
	if (globals->net.state != 2) {
		#ifdef DEBUG_NET_INFO
		printf("net_server(): not connected, clearing event queue\n");
		#endif
		event_queue_reset(&globals->net.events);
		return(1);
	}

pending_events:
	if (ret_val == 2) return(2);
	if (event_available(&globals->net.events) == 1) return(2);
	else return(1);

/* client has disconnected */
disconnected:
	if (globals->net.client != -1) close(globals->net.client);
	globals->net.client = -1;

	if (globals->net.server != -1) close(globals->net.server);
	globals->net.server = -1;

	globals->net.state = 0;
	event_queue_reset(&globals->net.events);
	
	/* listen for incoming connections */
	return(net_server(globals, NULL));
}

/**
 *  @brief if node is online notify user client (net)
 *
 *  This is a tree_walk callback.
 *
 *  @param node	pointer to the node to check
 *  @param data	pointer to globals
 *
 *  @returns 1
 */
static int net_notify_con_tree_walk(struct tree_node_t *node, void *data)
{
	struct btchat_node_t *n = NULL;
	struct btchatd_t *g = NULL;
	
	
	#ifdef DEBUG_NET
	assert(node != NULL);
	assert(data != NULL);
	#endif


	n = (struct btchat_node_t*) node->data;
	g = (struct btchatd_t*) data;

	return(bt_notify_net(g, n, 0));
}

/**
 *  @brief notify fresh connected user client about online nodes
 *
 *  @pram globals pointer to globals
 *
 *  @returns all ways 1
 */
int net_notify_con(struct btchatd_t *globals)
{
	void *old_walk;


	#ifdef DEBUG_NET
	assert(globals != NULL);
	#endif

	
	/* save old walk function */
	old_walk = globals->nodes.walk_function;
	/* install new walk function */
	globals->nodes.walk_function = net_notify_con_tree_walk;
	
	/* walk tree */
	tree_walk(&globals->nodes, globals);
	
	/* restore old walk function */
	globals->nodes.walk_function = old_walk;
	
	return(1);
}

/*
 *  $Log$
 */
