/*
 *  $Id$
 */

/**
 *  @file node_protocol.h
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

/*
 *  Note:
 *   all multi-byte values are in network-byte-order!
 *
 */

#ifndef __NODE_PROTOCOL_H__
#define __NODE_PROTOCOL_H__

enum btcnode_protocol_pkgtype_t {
	btcnode_protocol_null = 0,
	btcnode_protocol_trust,
	btcnode_protocol_route,
	btcnode_protocol_route_full,
	btcnode_protocol_msg,
	btcnode_protocol_channel,
	btcnode_protocol_mapreq = 253,
	btcnode_protocol_mapres,
	btcnode_protocol_unknown = 255
};

/**
 *  @brief protocol header
 *
 *  every protocol event starts with this header
 *
 */
struct btcnode_protocol_header_t {
	__uint8_t version;  /**< protocol version */
	__uint8_t pkg_type; /**< packet type - see btcnode_protocol_pkgtype_t */
	__uint32_t time;    /**< time at which packet was send (un*x time in seconds) */
	__uint16_t length;  /**< packet length, inclusive header */
};

struct btcnode_protocol_trust_t {
	__uint8_t route;   /**< routing level trust */
	__uint8_t general; /**< general level of trust - currently unused! */
};

/**
 *  @brief routing event
 *
 *  routing packages indicate that the given destination (dst)
 *  is reachable thru the sender of this package. One (1) or
 *  more destinations can be given.
 *
 *  pkg_type = btcnode_protocol_route
 *  length = sizeof(btcnode_protocol_header) + (sizeof(btcnode_protocol_route) * n)
 *  n = number of destinations included in the package (n >= 1)
 *
 */
struct btcnode_protocol_route_t {	
	__uint8_t dst[6]; /**< destination bd_addr */
	__uint8_t state;  /**< state */
	__uint8_t hops;   /**< number of hops (255 = not available anymore) */
};

/**
 *  @brief full routing event
 *
 *  difference to "normal" routing event is that this one
 *  contains the name of the node, this full routing event is
 *  when a device initial connects to the net
 *
 */
struct btcnode_protocol_route_full_t {	
	__uint8_t dst[6];    /**< destination bd_addr */
	__uint8_t state;     /**< state */
	__uint8_t hops;      /**< number of hops (255 = not available anymore) */
	__uint8_t name[250]; /**< name */
	__uint16_t reserved; /**< reserved */
};

/**
 *  @brief point-to-point message
 *
 *  pkg_type = btcnode_protocol_msg
 *  length = sizeof(btchat_protocol_header) + sizeof(btcnode_protocol_msg) + strlen(msg)
 *
 */
struct btcnode_protocol_msg_t {
	__uint8_t src[6]; /**< source bdaddr */
	__uint8_t dst[6]; /**< destination bdaddr */
	
	/* __uint8_t *msg; */   /**< message bytes */
};

/**
 *  @brief channel/broadcast message
 *
 *  pkg_type = btcnode_protocol_channel
 *  length = sizeof(btchat_protocol_header) + sizeof(btcnode_protocol_channel) + strlen(msg)
 *
 */
struct btcnode_protocol_channel_t {
	__uint8_t src[6];       /**< source bdaddr */
	__uint8_t dst[6];       /**< destination bdaddr, normally FF:FF:FF:FF:FF:FF */
	__uint8_t channel[255]; /**< name of channel */
	/* __uint8_t *msg; */   /**< message bytes */
};

/**
 *  @brief map request
 *
 *  pkg_type = btcnode_protocol_mapreq
 *  length = sizeof(btchat_protocol_header) + sizeof(btcnode_protocol_mapreq)
 *
 */
struct btcnode_protocol_mapreq_t {
	__uint8_t src[6];   /**< source bdaddr */
	__uint8_t dst[6];   /**< destination bdaddr, normally FF:FF:FF:FF:FF:FF */
	__uint32_t session; /**< session id - used in answer */
};

/**
 *  @brief map responds
 *
 *  pkg_type = btcnode_protocol_mapres
 *  length = sizeof(btchat_protocol_header) + sizeof(btcnode_protocol_mapres)
 *
 */
struct btcnode_protocol_mapres_t {
	__uint8_t src[6];    /**< source bdaddr */
	__uint8_t dst[6];    /**< destination bdaddr */
	__uint32_t session;  /**< session id - from request */
	__uint8_t name[255]; /**< name of node */
	struct {
		__uint8_t addr[6];   /**< node addr */
		__uint8_t direction; /**< incoming or outgoing */
		__uint32_t send;     /**< bytes send */
		__uint32_t recv;     /**< bytes received */
		__uint8_t state;     /**< role of link 0=master, 1=slave */
	} neigbours[7];
};

#endif

/*
 *  $Log$
 */
