/*
 *  $Id$
 */
  
/**
 *  @file sdp.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include "tree.h"
#include "btchatd.h"
#include "net.h"
#include "bt.h"
#include "util.h"
#include "discover_protocol.h"
#include "node_protocol.h"
#include "event.h"

/**
 *  @brief register btChat at the local SDP server
 *
 *  @param globals pointer to btchatd globals
 *
 *  @returns -1 on error and 1 on success
 */
int bt_sdp_register(struct btchatd_t *globals)
{
	sdp_list_t *svclass, *pfseq, *apseq, *root, *aproto;
	uuid_t root_uuid, l2cap, rfcomm, btchat;
	sdp_profile_desc_t profile[1];
	sdp_list_t *proto[2];
	int status;
	
	
	globals->btserver.sdp_session = sdp_connect(BDADDR_ANY, BDADDR_LOCAL, 0);
	if (!globals->btserver.sdp_session) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_register(): sdp_connect() failed\n");
		#endif
		return(-1);
	}
	
	globals->btserver.sdp_record = sdp_record_alloc();
	if (!globals->btserver.sdp_record) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_register(): sdp_record_alloc() failed\n");
		#endif
		return(-1);
	}
	
	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
	root = sdp_list_append(NULL, &root_uuid);
	sdp_set_browse_groups(globals->btserver.sdp_record, root);
	
	sdp_uuid16_create(&l2cap, L2CAP_UUID);
	proto[0] = sdp_list_append(NULL, &l2cap);
	apseq = sdp_list_append(NULL, proto[0]);
	
	sdp_uuid16_create(&rfcomm, RFCOMM_UUID);
	proto[1] = sdp_list_append(NULL, &rfcomm);
	proto[1] = sdp_list_append(proto[1], sdp_data_alloc(SDP_UINT8, &globals->self.btchat_channel));
	apseq = sdp_list_append(apseq, proto[1]);
	
	aproto = sdp_list_append(NULL, apseq);
	sdp_set_access_protos(globals->btserver.sdp_record, aproto);
	
	sdp_uuid16_create(&btchat, 0x2222);
	svclass = sdp_list_append(NULL, &btchat);
	sdp_set_service_classes(globals->btserver.sdp_record, svclass);
	
	sdp_uuid16_create(&profile[0].uuid, 0x2222);
	profile[0].version = 0x0100;
	pfseq = sdp_list_append(NULL, &profile[0]);
	sdp_set_profile_descs(globals->btserver.sdp_record, pfseq);
	
	sdp_set_info_attr(globals->btserver.sdp_record, "Collin's btChat protocol", NULL, NULL);

	status = sdp_record_register(globals->btserver.sdp_session, globals->btserver.sdp_record, 0);
	if (status) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_register(): sdp_record_register() failed\n");
		#endif
		sdp_record_free(globals->btserver.sdp_record);
		globals->btserver.sdp_record = NULL;
		return(-1);
	}

	return(1);
}

/**
 *  @brief unregister btChat at the local SDP server
 *
 *  @param globals pointer to btchatd globals
 *
 *  @returns 1
 */
int bt_sdp_unregister(struct btchatd_t *globals)
{
	if (globals->btserver.sdp_record && sdp_record_unregister(globals->btserver.sdp_session, globals->btserver.sdp_record)) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_unregister(): sdp_record_unregister() failed\n");
		#endif
	}
	
	sdp_close(globals->btserver.sdp_session);
	
	return(1);
}

/**
 *  @brief search for btChat at the given BD_ADDR
 *
 *  @param src     pointer to source (local) bd_addr
 *  @param dst     pointer to destination (remote) bd_addr
 *  @param channel pointer where to put found channel
 *
 *  @returns -1 on error 0 when not found and 1 when found
 */
int bt_sdp_search(bdaddr_t *src, bdaddr_t *dst, int *channel)
{
	sdp_session_t *session;
	sdp_list_t *srch, *attrs, *rsp;
	uuid_t svclass;
	uint16_t attr;
	int err;
	int ch;


	session = sdp_connect(src, dst, 0);
	if (!session) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_search(): sdp_connect() failed\n");
		#endif
		return(-1);
	}
	
	sdp_uuid16_create(&svclass, 0x2222);
	srch  = sdp_list_append(NULL, &svclass);
	
	attr  = SDP_ATTR_PROTO_DESC_LIST;
	attrs = sdp_list_append(NULL, &attr);
	
	err = sdp_service_search_attr_req(session, srch, SDP_ATTR_REQ_INDIVIDUAL, attrs, &rsp);
	
	sdp_close(session);

	if (err) {
		#ifdef DEBUG_SDP
		printf("bt_sdp_search(): sdp_service_search_attr_req() failed\n");
		#endif
		return(0);
	}
	
	for(; rsp; rsp = rsp->next) {
		sdp_record_t *rec = (sdp_record_t *) rsp->data;
		sdp_list_t *protos;
		
		if (!sdp_get_access_protos(rec, &protos)) {
			ch = sdp_get_proto_port(protos, RFCOMM_UUID);
			if (ch > 0) {
				*channel = ch;
				return(1);
			}
		}
	}
	
	return(0);
}

/*
 *  $Log$
 */
