/*
 *  $Id$
 */

/**
 *  @file tree.c
 *
 *  @project btChat (http://www.mulliner.org/bluetooth/btchat/)
 *
 *  @author Collin R. Mulliner <collin(AT)betaversion.net>
 *
 *  (c) Collin R. Mulliner
 *
 */

#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include "tree.h"

static void tree_rotate_left(struct tree_node_t *node)
{
	struct tree_node_t *tmp;
	struct tree_node_t *gparent;
	struct tree_node_t *parent = node->p;
	
	
	if (parent->p != NULL) {
		gparent = parent->p;
		if (gparent->l == parent) gparent->l = node;
		else gparent->r = node;
	}
	
	tmp = node->r;
	node->r = parent;
	if (tmp != NULL) tmp->p = parent;
	parent->l = tmp;
	node->p = parent->p;
	parent->p = node;
}

static void tree_rotate_right(struct tree_node_t *node)
{
	struct tree_node_t *tmp;
	struct tree_node_t *gparent;
	struct tree_node_t *parent = node->p;
	
	
	if (parent->p != NULL) {
		gparent = parent->p;
		if (gparent->l == parent) gparent->l = node;
		else gparent->r = node;
	}
	
	tmp = node->l;
	node->l = parent;
	if (tmp != NULL) tmp->p = parent;
	parent->r = tmp;
	node->p = parent->p;
	parent->p = node;
}

static void tree_splay(struct tree_node_t *node)
{
	struct tree_node_t *p;
	struct tree_node_t *g;
	
	
	if (node != NULL) {
		while (1) {
			p = node->p;
			if (p == NULL) break;
			g = p->p;
			
			if (g == NULL) {
				if (p->l == node) {
					tree_rotate_left(node);
					break;
				}
				else {
					tree_rotate_right(node);
					break;
				}
			}
			else {
				if (g->l == p) {
					if (p->l == node) {
						tree_rotate_left(p);
						tree_rotate_left(node);
					}
					else {
						tree_rotate_right(node);
						tree_rotate_left(node);
					}
				}
				else {
					if (p->r == node) {
						tree_rotate_right(p);
						tree_rotate_right(node);
					}
					else {
						tree_rotate_left(node);
						tree_rotate_right(node);
					}
				}
			}
		} /* while(1) */
	} /* if (node != NULL) */
}

static int tree_walk_search(struct tree_node_t *cur, tree_walk_callback *walk, void *user)
{
	/* go left */
	if (cur->l != NULL)
		/* 2 == interrupt - break walk */
		if (tree_walk_search(cur->l, walk, user) == 2) return(2);
		
	/* go right */
	if (cur->r != NULL)
		/* 2 == interrupt - break walk */
		if (tree_walk_search(cur->r, walk, user) == 2) return(2);
	
	/* do user action for current node */
	return(walk(cur, user));
}

/**
 *  @brief tree traversal
 *
 *  @param tree pointer to tree
 *  @param user pointer to user data (is handed over to walk function)
 *
 *  @returns value from walk function
 */
int tree_walk(const struct tree_t *tree, void *user)
{
	if (tree->root == NULL) return(1);

	return(tree_walk_search(tree->root, tree->walk_function, user));
}

static int tree_reset_walk(struct tree_node_t **cur, int delete_data)
{
	if (cur == NULL) return(0);
	if (*cur == NULL) return(0);

	tree_reset_walk(&((*cur)->l), delete_data);
	tree_reset_walk(&((*cur)->r), delete_data);
	
	if (delete_data == 1) {
		free((*cur)->data);
		(*cur)->data = NULL;
	}
	free(*cur);
	*cur = NULL;
	
	return(1);

}

/**
 *  @brief reset/delete tree
 *
 *  @param tree        pointer to tree
 *  @param delete_data also free data (1=yes, 0=no)
 *
 *  @returns 1
 */
int tree_reset(struct tree_t *tree, int delete_data)
{
	tree->num = 0;
	return(tree_reset_walk(&tree->root, delete_data));
}

static int tree_access_search(struct tree_node_t *cur, const void *search, struct tree_node_t **result, tree_compare_callback *cmp)
{
	int res;

	
	#ifdef DEBUG
	assert(search != NULL);
	assert(result != NULL);
	assert(cmp != NULL);
	#endif
	

	if (cur != NULL) {
		res = cmp(cur->data, search);
		
		if (res < 0) {
			if (cur->l == NULL) {
				*result = cur;
				return(0);
			}
			else {
				return(tree_access_search(cur->l, search, result, cmp));
			}
		}
		else if (res > 0) {
			if (cur->r == NULL) {
				*result = cur;
				return(1);
			}
			else {
				return(tree_access_search(cur->r, search, result, cmp));
			}
		}
		else {
			*result = cur;
			return(2);
		}
	}
	else {
		return(-1);
	}
}

/**
 *  @brief remove given node from tree
 *
 *  @param tree        pointer to tree
 *  @param search      pointer to tree data element
 *  @param delete_data delete/free node data, 1=yes and 0=no
 *
 *  @returns 1 on success and 0 on failure
 */
int tree_remove(struct tree_t *tree, const void *search, int delete_data)
{
	struct tree_node_t *result = NULL;
	struct tree_node_t *l;
	struct tree_node_t *r;
	struct tree_node_t *p;
	int res;
	
	
	#ifdef DEBUG
	assert(tree != NULL);
	assert(search != NULL);
	#endif
	
	
	res = tree_access(tree, search, &result);
	
	if (res == 1) {
		l = result->l;
		r = result->r;
		
		if (l != NULL) {
			l->p = NULL;
			
			while (1) {
				if (l->r != NULL) {
					l = l->r;
				}
				else {
					tree_splay(l);
					l->r = r;
					if (r != NULL) r->p = l;
					break;
				}
			}
		}
		else {
			if (r != NULL) r->p = NULL;
			l = r;
		}
		
		if (result->p != NULL) {
			p = result->p;
			if (p->l == result) p->l = l;
			else p->r = l;
			if (l != NULL) l->p = p;
			tree_splay(p);
			l = p;
		}
		
		if (delete_data == 1) {
			if (result->data != NULL) free(result->data);
		}
		#ifdef DEBUG
		assert(result != NULL);
		#endif
		free(result);
		tree->root = l;
		
		return(1);
	}
	
	return(0);
}

/**
 *  @brief find search element in tree
 *
 *  the result is written to result
 *
 *  @param tree   pointer to tree
 *  @param search pointer to search data (interpreted by compare function)
 *  @param result pointer to pointer for result
 *
 *  @returns -1 on parameter error, 0 if search failed and 1 on success
 */
int tree_access(struct tree_t *tree, const void *search, struct tree_node_t **result)
{
	int res;
	

	if (tree == NULL) return(-1);
	if (search == NULL) return(-1);
	if (result == NULL) return(-1);

	if (tree->root == NULL) return(0);

	res = tree_access_search(tree->root, search, result, tree->compare_function);
	
	/* use splaying for optimization */
	if (tree->splay == 1) {
		tree_splay(*result);
		tree->root = *result;
	}
	
	switch (res) {
	case 2:
		return(1);
	case -1:
		return(-1);
	default:
		return(0);
	}
}

/**
 *  @brief find search element in tree or create new element
 *
 *  The result is written to result, when the search fails (with result == 0)
 *  the new allocated node is return as result.
 *
 *  @param tree   pointer to tree
 *  @param search pointer to search data (interpreted by compare function)
 *  @param result pointer to pointer for result
 *
 *  @returns -1 on parameter error, 0 if new node was allocated and 1 on success
 */
int tree_access_insert(struct tree_t *tree, const void *search, struct tree_node_t **result)
{
	int res;
	struct tree_node_t *tmp;
	

	if (tree == NULL) return(-1);
	if (search == NULL) return(-1);
	if (result == NULL) return(-1);

	if (tree->root == NULL) {
		tree->root = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		memset(tree->root, 0, sizeof(struct tree_node_t));
		tree->num = 1;
		*result = tree->root;
		return(0);
	}

	res = tree_access_search(tree->root, search, result, tree->compare_function);

	tmp = *result;
	if (res == 0) {
		(*result)->l = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		*result = (*result)->l;
	}
	else if (res == 1) {
		(*result)->r = (struct tree_node_t*) malloc(sizeof(struct tree_node_t));
		*result = (*result)->r;
	}
	else {
		return((res == 2 ? 1 : res));
	}
	
	memset((*result), 0, sizeof(struct tree_node_t));
	(*result)->p = tmp;
	tree->num++;
	
	/* use splaying for optimization */
	if (tree->splay == 1) {
		tree_splay(*result);
		tree->root = *result;
	}
	
	return(0);
}

/*
 *  $Log$
 */
